@@ -22,9 +22,55 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <unistd.h>
 #include "automount.h"
 #include "nsswitch.h"
 
+static int check_nss_result(struct nss_source *this, enum nsswitch_status result)
+{
+	enum nsswitch_status status;
+	struct nss_action a;
+
+	/* Check if we have negated actions */
+	for (status = 0; status < NSS_STATUS_MAX; status++) {
+		a = this->action[status];
+		if (a.action == NSS_ACTION_UNKNOWN)
+			continue;
+
+		if (a.negated && result != status) {
+			if (a.action == NSS_ACTION_RETURN) {
+				if (result == NSS_STATUS_SUCCESS)
+					return 1;
+				else
+					return 0;
+			}
+		}
+	}
+
+	a = this->action[result];
+
+	/* Check if we have other actions for this status */
+	switch (result) {
+	case NSS_STATUS_SUCCESS:
+		if (a.action == NSS_ACTION_CONTINUE)
+			break;
+		return 1;
+
+	case NSS_STATUS_NOTFOUND:
+	case NSS_STATUS_UNAVAIL:
+	case NSS_STATUS_TRYAGAIN:
+		if (a.action == NSS_ACTION_RETURN) {
+			return 0;
+		}
+		break;
+
+	default:
+		break;
+	}
+
+	return -1;
+}
+
 static void nsslist_cleanup(void *arg)
 {
 	struct list_head *nsslist = (struct list_head *) arg;
@@ -44,9 +90,9 @@ static int do_read_master(struct master *master, char *type, time_t age)
 	argv[0] = master->name;
 	argv[1] = NULL;
 
-	status = open_lookup(type, "", NULL, argc, argv, &lookup);
-	if (status != NSS_STATUS_SUCCESS)
-		return status;
+	lookup = open_lookup(type, "", NULL, argc, argv);
+	if (!lookup)
+		return NSS_STATUS_UNAVAIL;
 
 	status = lookup->lookup_read_master(master, age, lookup->context);
 
@@ -55,66 +101,6 @@ static int do_read_master(struct master *master, char *type, time_t age)
 	return status;
 }
 
-static char *find_map_path(struct autofs_point *ap, struct map_source *map)
-{
-	const char *mname = map->argv[0];
-	unsigned int mlen = strlen(mname);
-	char *tok, *ptr = NULL;
-	char *path = NULL;
-	char *search_path;
-	struct stat st;
-
-	/*
-	 * This is different to the way it is in amd.
-	 * autofs will always try to locate maps in AUTOFS_MAP_DIR
-	 * but amd has no default and will not find a file map that
-	 * isn't a full path when no search_path is configured, either
-	 * in the mount point or global configuration.
-	 */
-	search_path = strdup(AUTOFS_MAP_DIR);
-	if (map->flags & MAP_FLAG_FORMAT_AMD) {
-		struct autofs_point *pap = ap;
-		char *tmp;
-		/*
-		 * Make sure we get search_path from the root of the
-		 * mount tree, if one is present in the configuration.
-		 * Again different from amd, which ignores the submount
-		 * case.
-		 */
-		while (pap->parent)
-			pap = pap->parent;
-		tmp = conf_amd_get_search_path(pap->path);
-		if (tmp) {
-			if (search_path)
-				free(search_path);
-			search_path = tmp;
-		}
-	}
-	if (!search_path)
-		return NULL;
-
-	tok = strtok_r(search_path, ":", &ptr);
-	while (tok) {
-		char *this = malloc(strlen(tok) + mlen + 2);
-		if (!this) {
-			free(search_path);
-			return NULL;
-		}
-		strcpy(this, tok);
-		strcat(this, "/");
-		strcat(this, mname);
-		if (!stat(this, &st)) {
-			path = this;
-			break;
-		}
-		free(this);
-		tok = strtok_r(NULL, ":", &ptr);
-	}
-
-	free(search_path);
-	return path;
-}
-
 static int read_master_map(struct master *master, char *type, time_t age)
 {
 	unsigned int logopt = master->logopt;
@@ -172,10 +158,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			result = do_read_master(master, "file", age);
 		}
 
-		if (result == NSS_STATUS_UNAVAIL)
-			master->read_fail = 1;
-
-		return result;
+		return !result;
 	} else {
 		char *name = master->name;
 		char *tmp;
@@ -186,22 +169,12 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			char source[10];
 
 			memset(source, 0, 10);
-			if ((!strncmp(name, "file", 4) &&
-				 (name[4] == ',' || name[4] == ':')) ||
-			    (!strncmp(name, "yp", 2) &&
-				 (name[2] == ',' || name[2] == ':')) ||
-			    (!strncmp(name, "nis", 3) &&
-				 (name[3] == ',' || name[3] == ':')) ||
-			    (!strncmp(name, "nisplus", 7) &&
-				 (name[7] == ',' || name[7] == ':')) ||
-			    (!strncmp(name, "ldap", 4) &&
-				 (name[4] == ',' || name[4] == ':')) ||
-			    (!strncmp(name, "ldaps", 5) &&
-				 (name[5] == ',' || name[5] == ':')) ||
-			    (!strncmp(name, "sss", 3) ||
-				 (name[3] == ',' || name[3] == ':')) ||
-			    (!strncmp(name, "dir", 3) &&
-				 (name[3] == ',' || name[3] == ':'))) {
+			if (!strncmp(name, "file:", 5) ||
+			    !strncmp(name, "yp:", 3) ||
+			    !strncmp(name, "nis:", 4) ||
+			    !strncmp(name, "nisplus:", 8) ||
+			    !strncmp(name, "ldap:", 5) ||
+			    !strncmp(name, "ldaps:", 6)) {
 				strncpy(source, name, tmp - name);
 
 				/*
@@ -222,10 +195,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 				result = do_read_master(master, source, age);
 				master->name = name;
 
-				if (result == NSS_STATUS_UNAVAIL)
-					master->read_fail = 1;
-
-				return result;
+				return !result;
 			}
 		}
 	}
@@ -237,11 +207,10 @@ int lookup_nss_read_master(struct master *master, time_t age)
 		if (!list_empty(&nsslist))
 			free_sources(&nsslist);
 		error(logopt, "can't to read name service switch config.");
-		return NSS_STATUS_UNAVAIL;
+		return 0;
 	}
 
 	/* First one gets it */
-	result = NSS_STATUS_SUCCESS;
 	head = &nsslist;
 	list_for_each(p, head) {
 		struct nss_source *this;
@@ -249,13 +218,6 @@ int lookup_nss_read_master(struct master *master, time_t age)
 
 		this = list_entry(p, struct nss_source, list);
 
-		if (strncmp(this->source, "files", 5) &&
-		    strncmp(this->source, "nis", 3) &&
-		    strncmp(this->source, "nisplus", 7) &&
-		    strncmp(this->source, "ldap", 4) &&
-		    strncmp(this->source, "sss", 3))
-			continue;
-
 		debug(logopt,
 		      "reading master %s %s", this->source, master->name);
 
@@ -282,22 +244,11 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			}
 		}
 
-		/* We've been instructed to move onto the next source */
-		if (result == NSS_STATUS_TRYAGAIN) {
-			result = NSS_STATUS_SUCCESS;
-			continue;
-		}
-
-		if (result == NSS_STATUS_UNKNOWN ||
-		    result == NSS_STATUS_NOTFOUND) {
+		if (result == NSS_STATUS_UNKNOWN) {
 			debug(logopt, "no map - continuing to next source");
-			result = NSS_STATUS_SUCCESS;
 			continue;
 		}
 
-		if (result == NSS_STATUS_UNAVAIL)
-			master->read_fail = 1;
-
 		status = check_nss_result(this, result);
 		if (status >= 0) {
 			free_sources(&nsslist);
@@ -308,7 +259,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 	if (!list_empty(&nsslist))
 		free_sources(&nsslist);
 
-	return result;
+	return !result;
 }
 
 static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t age)
@@ -316,29 +267,29 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 	struct lookup_mod *lookup;
 	int status;
 
-	lookup = NULL;
-	master_source_writelock(ap->entry);
-	if (!map->lookup) {
-		status = open_lookup(map->type, "", map->format,
-				     map->argc, map->argv, &lookup);
-		if (status != NSS_STATUS_SUCCESS) {
-			master_source_unlock(ap->entry);
-			debug(ap->logopt,
-			      "lookup module %s open failed", map->type);
-			return status;
-		}
-		map->lookup = lookup;
-	} else {
-		lookup = map->lookup;
-		status = lookup->lookup_reinit(map->format,
-					       map->argc, map->argv,
-					       &lookup->context);
-		if (status)
-			warn(ap->logopt,
-			     "lookup module %s reinit failed", map->type);
+	lookup = open_lookup(map->type, "", map->format, map->argc, map->argv);
+	if (!lookup) {
+		debug(ap->logopt, "lookup module %s failed", map->type);
+		return NSS_STATUS_UNAVAIL;
 	}
+
+	master_source_writelock(ap->entry);
+	if (map->lookup)
+		close_lookup(map->lookup);
+	map->lookup = lookup;
 	master_source_unlock(ap->entry);
 
+	/* If we don't need to create directories then there's no use
+	 * reading the map. We just need to test that the map is valid
+	 * for the fail cases to function correctly and to cache the
+	 * lookup handle.
+	 *
+	 * We always need to whole map for direct mounts in order to
+	 * mount the triggers.
+	 */
+	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT)
+		return NSS_STATUS_SUCCESS;
+
 	if (!map->stale)
 		return NSS_STATUS_SUCCESS;
 
@@ -347,20 +298,14 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 
 	status = lookup->lookup_read_map(ap, age, lookup->context);
 
-	if (status != NSS_STATUS_SUCCESS)
-		map->stale = 0;
+	map->stale = 0;
 
 	/*
 	 * For maps that don't support enumeration return success
 	 * and do whatever we must to have autofs function with an
 	 * empty map entry cache.
-	 *
-	 * For indirect maps that use the browse option, when the
-	 * server is unavailable continue as best we can with
-	 * whatever we have in the cache, if anything.
 	 */
-	if (status == NSS_STATUS_UNKNOWN ||
-	   (ap->type == LKP_INDIRECT && status == NSS_STATUS_UNAVAIL))
+	if (status == NSS_STATUS_UNKNOWN)
 		return NSS_STATUS_SUCCESS;
 
 	return status;
@@ -434,45 +379,6 @@ static void argv_cleanup(void *arg)
 	return;
 }
 
-static int lookup_map_read_map(struct autofs_point *ap,
-			       struct map_source *map, time_t age)
-{
-	char *path;
-
-	if (!map->argv[0]) {
-		if (!strcmp(map->type, "hosts"))
-			return do_read_map(ap, map, age);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	/*
-	 * This is only called when map->type != NULL.
-	 * We only need to look for a map if source type is
-	 * file and the map name doesn't begin with a "/".
-	 */
-	if (strncmp(map->type, "file", 4))
-		return do_read_map(ap, map, age);
-
-	if (map->argv[0][0] == '/')
-		return do_read_map(ap, map, age);
-
-	path = find_map_path(ap, map);
-	if (!path)
-		return NSS_STATUS_UNKNOWN;
-
-	if (map->argc >= 1) {
-		if (map->argv[0])
-			free((char *) map->argv[0]);
-		map->argv[0] = path;
-	} else {
-		error(ap->logopt, "invalid arguments for autofs_point");
-		free(path);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	return do_read_map(ap, map, age);
-}
-
 static enum nsswitch_status read_map_source(struct nss_source *this,
 		struct autofs_point *ap, struct map_source *map, time_t age)
 {
@@ -497,24 +403,24 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 	}
 
 	this->source[4] = '\0';
-	tmap.flags = map->flags;
 	tmap.type = this->source;
 	tmap.format = map->format;
-	tmap.name = map->name;
 	tmap.lookup = map->lookup;
 	tmap.mc = map->mc;
 	tmap.instance = map->instance;
-	tmap.exp_timeout = map->exp_timeout;
 	tmap.recurse = map->recurse;
 	tmap.depth = map->depth;
-	tmap.stale = map->stale;
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
-	path = find_map_path(ap, map);
+	path = malloc(strlen(AUTOFS_MAP_DIR) + strlen(map->argv[0]) + 2);
 	if (!path)
 		return NSS_STATUS_UNKNOWN;
 
+	strcpy(path, AUTOFS_MAP_DIR);
+	strcat(path, "/");
+	strcat(path, map->argv[0]);
+
 	if (map->argc >= 1) {
 		tmap.argc = map->argc;
 		tmap.argv = copy_argv(map->argc, map->argv);
@@ -536,8 +442,7 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 	result = read_file_source_instance(ap, &tmap, age);
 	pthread_cleanup_pop(1);
 
-	if (!map->instance)
-		map->instance = tmap.instance;
+	map->instance = tmap.instance;
 
 	return result;
 }
@@ -576,7 +481,7 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 				debug(ap->logopt,
 				      "reading map %s %s",
 				       map->type, map->argv[0]);
-			result = lookup_map_read_map(ap, map, age);
+			result = do_read_map(ap, map, age);
 			map = map->next;
 			continue;
 		}
@@ -619,13 +524,6 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 		list_for_each(p, head) {
 			this = list_entry(p, struct nss_source, list);
 
-			if (map->flags & MAP_FLAG_FORMAT_AMD &&
-			    !strcmp(this->source, "sss")) {
-				warn(ap->logopt,
-				     "source sss is not available for amd maps.");
-				continue;
-			}
-
 			debug(ap->logopt,
 			      "reading map %s %s", this->source, map->argv[0]);
 
@@ -633,8 +531,7 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 			if (result == NSS_STATUS_UNKNOWN)
 				continue;
 
-			/* Try to avoid updating the map cache if an instance
-			 * is unavailable */
+			/* Don't try to update the map cache if it's unavailable */
 			if (result == NSS_STATUS_UNAVAIL)
 				map->stale = 0;
 
@@ -665,56 +562,6 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 	return 0;
 }
 
-static char *make_browse_path(unsigned int logopt,
-			      const char *root, const char *key,
-			      const char *prefix)
-{
-	unsigned int l_prefix;
-	unsigned int k_len, r_len;
-	char *k_start;
-	char *path;
-
-	k_start = (char *) key;
-	k_len = strlen(key);
-	l_prefix = 0;
-
-	if (prefix) {
-		l_prefix = strlen(prefix);
-
-		if (l_prefix > k_len)
-			return NULL;
-
-		/* If the prefix doesn't match the beginning
-		 * of the key this entry isn't a sub directory
-		 * at this level.
-		 */
-		if (strncmp(key, prefix, l_prefix))
-			return NULL;
-
-		/* Directory entry starts following the prefix */
-		k_start += l_prefix;
-	}
-
-	/* No remaining "/" allowed here */
-	if (strchr(k_start, '/'))
-		return NULL;
-
-	r_len = strlen(root);
-
-	if ((r_len + strlen(k_start)) > KEY_MAX_LEN)
-		return NULL;
-
-	path = malloc(r_len + k_len + 2);
-	if (!path) {
-		warn(logopt, "failed to allocate full path");
-		return NULL;
-	}
-
-	sprintf(path, "%s/%s", root, k_start);
-
-	return path;
-}
-
 int lookup_ghost(struct autofs_point *ap, const char *root)
 {
 	struct master_mapent *entry = ap->entry;
@@ -750,27 +597,10 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 		cache_readlock(mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
-			/*
-			 * Map entries that have been created in the cache
-			 * due to a negative lookup shouldn't have directories
-			 * created if they haven't already been created.
-			 */
-			if (!me->mapent)
+			if (*me->key == '*')
 				goto next;
 
-			/* Wildcard cannot be a browse directory and amd map
-			 * keys may end with the wildcard.
-			 */
-			if (strchr(me->key, '*'))
-				goto next;
-
-			/* This will also take care of amd "/defaults" entry as
-			 * amd map keys are not allowd to start with "/"
-			 */
 			if (*me->key == '/') {
-				if (map->flags & MAP_FLAG_FORMAT_AMD)
-					goto next;
-
 				/* It's a busy multi-mount - leave till next time */
 				if (list_empty(&me->multi_list))
 					error(ap->logopt,
@@ -778,10 +608,12 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 				goto next;
 			}
 
-			fullpath = make_browse_path(ap->logopt,
-						    root, me->key, ap->pref);
-			if (!fullpath)
+			fullpath = malloc(strlen(me->key) + strlen(root) + 3);
+			if (!fullpath) {
+				warn(ap->logopt, "failed to allocate full path");
 				goto next;
+			}
+			sprintf(fullpath, "%s/%s", root, me->key);
 
 			ret = stat(fullpath, &st);
 			if (ret == -1 && errno != ENOENT) {
@@ -791,18 +623,7 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 				goto next;
 			}
 
-			/* Directory already exists? */
-			if (!ret) {
-				/* Shouldn't need this
-				me->dev = st.st_dev;
-				me->ino = st.st_ino;
-				*/
-				debug(ap->logopt, "me->dev %d me->ino %d", me->dev, me->ino);
-				free(fullpath);
-				goto next;
-			}
-
-			ret = mkdir_path(fullpath, mp_mode);
+			ret = mkdir_path(fullpath, 0555);
 			if (ret < 0 && errno != EEXIST) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				warn(ap->logopt,
@@ -834,12 +655,12 @@ int do_lookup_mount(struct autofs_point *ap, struct map_source *map, const char
 	int status;
 
 	if (!map->lookup) {
-		status = open_lookup(map->type, "",
-				     map->format, map->argc, map->argv, &lookup);
-		if (status != NSS_STATUS_SUCCESS) {
+		lookup = open_lookup(map->type, "",
+				     map->format, map->argc, map->argv);
+		if (!lookup) {
 			debug(ap->logopt,
-			      "lookup module %s open failed", map->type);
-			return status;
+			      "lookup module %s failed", map->type);
+			return NSS_STATUS_UNAVAIL;
 		}
 		map->lookup = lookup;
 	}
@@ -854,88 +675,17 @@ int do_lookup_mount(struct autofs_point *ap, struct map_source *map, const char
 	return status;
 }
 
-static int lookup_amd_instance(struct autofs_point *ap,
-			       struct map_source *map,
-			       const char *name, int name_len)
-{
-	struct map_source *instance;
-	struct amd_entry *entry;
-	const char *argv[2];
-	const char **pargv = NULL;
-	int argc = 0;
-	struct mapent *me;
-	char *m_key;
-
-	me = cache_lookup_distinct(map->mc, name);
-	if (!me || !me->multi) {
-		error(ap->logopt, "expected multi mount entry not found");
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	m_key = malloc(strlen(ap->path) + strlen(me->multi->key) + 2);
-	if (!m_key) {
-		error(ap->logopt,
-		     "failed to allocate storage for search key");
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	strcpy(m_key, ap->path);
-	strcat(m_key, "/");
-	strcat(m_key, me->multi->key);
-	entry = master_find_amdmount(ap, m_key);
-	free(m_key);
-
-	if (!entry) {
-		error(ap->logopt, "expected amd mount entry not found");
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	if (strcmp(entry->type, "host")) {
-		error(ap->logopt, "unexpected map type %s", entry->type);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	if (entry->opts && *entry->opts) {
-		argv[0] = entry->opts;
-		argv[1] = NULL;
-		pargv = argv;
-		argc = 1;
-	}
-
-	instance = master_find_source_instance(map, "hosts", "sun", argc, pargv);
-	/* If this is an nss map instance it may have an amd host map sub instance */
-	if (!instance && map->instance) {
-		struct map_source *next = map->instance;
-		while (next) {
-			instance = master_find_source_instance(next,
-						"hosts", "sun", argc, pargv);
-			if (instance)
-				break;
-			next = next->next;
-		}
-	}
-	if (!instance) {
-		error(ap->logopt, "expected hosts map instance not found");
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	return do_lookup_mount(ap, instance, name, name_len);
-}
-
 static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_source *map, const char *name, int name_len)
 {
 	struct map_source *instance;
 	char src_file[] = "file";
 	char src_prog[] = "program";
-	time_t age = monotonic_time(NULL);
+	time_t age = time(NULL);
 	struct stat st;
 	char *type, *format;
 
-	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
-		return lookup_amd_instance(ap, map, name, name_len);
-
 	if (stat(map->argv[0], &st) == -1) {
-		debug(ap->logopt, "file map not found");
+		warn(ap->logopt, "file map not found");
 		return NSS_STATUS_NOTFOUND;
 	}
 
@@ -967,10 +717,7 @@ static int lookup_name_source_instance(struct autofs_point *ap, struct map_sourc
 {
 	struct map_source *instance;
 	const char *format;
-	time_t age = monotonic_time(NULL);
-
-	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
-		return lookup_amd_instance(ap, map, name, name_len);
+	time_t age = time(NULL);
 
 	format = map->format;
 
@@ -988,49 +735,6 @@ static int lookup_name_source_instance(struct autofs_point *ap, struct map_sourc
 	return do_lookup_mount(ap, instance, name, name_len);
 }
 
-static int do_name_lookup_mount(struct autofs_point *ap,
-				struct map_source *map,
-				const char *name, int name_len)
-{
-	char *path;
-
-	if (!map->argv[0]) {
-		if (!strcmp(map->type, "hosts"))
-			return do_lookup_mount(ap, map, name, name_len);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
-		return lookup_amd_instance(ap, map, name, name_len);
-
-	/*
-	 * This is only called when map->type != NULL.
-	 * We only need to look for a map if source type is
-	 * file and the map name doesn't begin with a "/".
-	 */
-	if (strncmp(map->type, "file", 4))
-		return do_lookup_mount(ap, map, name, name_len);
-
-	if (map->argv[0][0] == '/')
-		return do_lookup_mount(ap, map, name, name_len);
-
-	path = find_map_path(ap, map);
-	if (!path)
-		return NSS_STATUS_UNKNOWN;
-
-	if (map->argc >= 1) {
-		if (map->argv[0])
-			free((char *) map->argv[0]);
-		map->argv[0] = path;
-	} else {
-		error(ap->logopt, "invalid arguments for autofs_point");
-		free(path);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	return do_lookup_mount(ap, map, name, name_len);
-}
-
 static enum nsswitch_status lookup_map_name(struct nss_source *this,
 			struct autofs_point *ap, struct map_source *map,
 			const char *name, int name_len)
@@ -1056,22 +760,23 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 	}
 
 	this->source[4] = '\0';
-	tmap.flags = map->flags;
 	tmap.type = this->source;
 	tmap.format = map->format;
-	tmap.name = map->name;
 	tmap.mc = map->mc;
 	tmap.instance = map->instance;
-	tmap.exp_timeout = map->exp_timeout;
 	tmap.recurse = map->recurse;
 	tmap.depth = map->depth;
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
-	path = find_map_path(ap, map);
+	path = malloc(strlen(AUTOFS_MAP_DIR) + strlen(map->argv[0]) + 2);
 	if (!path)
 		return NSS_STATUS_UNKNOWN;
 
+	strcpy(path, AUTOFS_MAP_DIR);
+	strcat(path, "/");
+	strcat(path, map->argv[0]);
+
 	if (map->argc >= 1) {
 		tmap.argc = map->argc;
 		tmap.argv = copy_argv(map->argc, map->argv);
@@ -1091,8 +796,7 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 
 	result = lookup_name_file_source_instance(ap, &tmap, name, name_len);
 
-	if (!map->instance)
-		map->instance = tmap.instance;
+	map->instance = tmap.instance;
 
 	/* path is freed in free_argv */
 	free_argv(tmap.argc, tmap.argv);
@@ -1100,99 +804,6 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 	return result;
 }
 
-static struct map_source *lookup_get_map_source(struct master_mapent *entry)
-{
-	struct map_source *map = entry->maps;
-	struct stat st;
-	char *type;
-
-	if (map->type || *map->argv[0] != '/')
-		return map;
-
-	if (*(map->argv[0] + 1) == '/')
-		return map;
-
-	if (stat(map->argv[0], &st) == -1)
-		return NULL;
-
-	if (!S_ISREG(st.st_mode))
-		return NULL;
-
-	if (st.st_mode & __S_IEXEC)
-		type = "program";
-	else
-		type = "file";
-
-	/* This is a file source with a path starting with "/".
-	 * But file maps can be either plain text or executable
-	 * so they use a map instance and the actual map source
-	 * remains untouched.
-	 */
-	return master_find_source_instance(map, type, map->format, 0, NULL);
-}
-
-static void update_negative_cache(struct autofs_point *ap, struct map_source *source, const char *name)
-{
-	struct master_mapent *entry = ap->entry;
-	struct map_source *map;
-	struct mapent *me;
-
-	/* Don't update negative cache for included maps */ 
-	if (source && source->depth)
-		return;
-
-	/* Don't update the wildcard */
-	if (strlen(name) == 1 && *name == '*')
-		return;
-
-	/* Have we recorded the lookup fail for negative caching? */
-	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
-	if (me)
-		/*
-		 *  Already exists in the cache, the mount fail updates
-		 *  will update negative timeout status.
-		 */
-		cache_unlock(me->mc);
-	else {
-		if (!defaults_disable_not_found_message()) {
-			/* This really should be a warning but the original
-			 * request for this needed it to be unconditional.
-			 * That produces, IMHO, unnecessary noise in the log
-			 * so a configuration option has been added to provide
-			 * the ability to turn it off.
-			 */
-			logmsg("key \"%s\" not found in map source(s).", name);
-		}
-
-		/* Doesn't exist in any source, just add it somewhere.
-		 * Also take care to use the same map source used by
-		 * map reads and key lookups for the update.
-		 */
-		if (source)
-			map = source;
-		else
-			map = lookup_get_map_source(entry);
-		if (map) {
-			time_t now = monotonic_time(NULL);
-			int rv = CHE_FAIL;
-
-			cache_writelock(map->mc);
-			me = cache_lookup_distinct(map->mc, name);
-			if (me)
-				rv = cache_push_mapent(me, NULL);
-			else
-				rv = cache_update(map->mc, map, name, NULL, now);
-			if (rv != CHE_FAIL) {
-				me = cache_lookup_distinct(map->mc, name);
-				if (me)
-					me->status = now + ap->negative_timeout;
-			}
-			cache_unlock(map->mc);
-		}
-	}
-	return;
-}
-
 int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const char *name, int name_len)
 {
 	struct master_mapent *entry = ap->entry;
@@ -1201,7 +812,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 	struct nss_source *this;
 	struct map_source *map;
 	enum nsswitch_status status;
-	int result = NSS_STATUS_UNKNOWN;
+	int result = 0;
 
 	/*
 	 * For each map source (ie. each entry for the mount
@@ -1220,7 +831,6 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		 * the map entry was last updated.
 		 */
 		if (entry->age > map->age) {
-			status = NSS_STATUS_UNAVAIL;
 			map = map->next;
 			continue;
 		}
@@ -1228,7 +838,8 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		sched_yield();
 
 		if (map->type) {
-			result = do_name_lookup_mount(ap, map, name, name_len);
+			result = do_lookup_mount(ap, map, name, name_len);
+
 			if (result == NSS_STATUS_SUCCESS)
 				break;
 
@@ -1242,7 +853,6 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 				char *tmp = strdup("ldap");
 				if (!tmp) {
 					map = map->next;
-					status = NSS_STATUS_TRYAGAIN;
 					continue;
 				}
 				map->type = tmp;
@@ -1271,18 +881,12 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		list_for_each(p, head) {
 			this = list_entry(p, struct nss_source, list);
 
-			if (map->flags & MAP_FLAG_FORMAT_AMD &&
-			    !strcmp(this->source, "sss")) {
-				warn(ap->logopt,
-				     "source sss is not available for amd maps.");
-				result = NSS_STATUS_UNAVAIL;
-				continue;
-			}
-
 			result = lookup_map_name(this, ap, map, name, name_len);
 
-			if (result == NSS_STATUS_UNKNOWN)
+			if (result == NSS_STATUS_UNKNOWN) {
+				map = map->next;
 				continue;
+			}
 
 			status = check_nss_result(this, result);
 			if (status >= 0) {
@@ -1301,16 +905,11 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 	}
 	if (ap->state != ST_INIT)
 		send_map_update_request(ap);
-
-	/*
-	 * The last source lookup will return NSS_STATUS_NOTFOUND if the
-	 * map exits and the key has not been found but the map may also
-	 * not exist in which case the key is also not found.
-	 */
-	if (result == NSS_STATUS_NOTFOUND || result == NSS_STATUS_UNAVAIL)
-		update_negative_cache(ap, source, name);
 	pthread_cleanup_pop(1);
 
+	if (result == NSS_STATUS_NOTFOUND)
+		error(ap->logopt, "key \"%s\" not found in map.", name);
+
 	return !result;
 }
 
@@ -1346,147 +945,35 @@ void lookup_close_lookup(struct autofs_point *ap)
 	return;
 }
 
-static char *make_fullpath(struct autofs_point *ap, const char *key)
+static char *make_fullpath(const char *root, const char *key)
 {
-	char *path = NULL;
 	int l;
+	char *path;
 
-	if (*key != '/')
-		path = make_browse_path(ap->logopt, ap->path, key, ap->pref);
-	else {
+	if (*key == '/') {
 		l = strlen(key) + 1;
 		if (l > KEY_MAX_LEN)
-			goto out;
+			return NULL;
 		path = malloc(l);
-		if (!path)
-			goto out;
 		strcpy(path, key);
+	} else {
+		l = strlen(key) + 1 + strlen(root) + 1;
+		if (l > KEY_MAX_LEN)
+			return NULL;
+		path = malloc(l);
+		sprintf(path, "%s/%s", root, key);
 	}
-out:
 	return path;
 }
 
-void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, time_t age)
-{
-	struct mapent *me, *this;
-	char *path;
-	int status = CHE_FAIL;
-
-	me = cache_enumerate(mc, NULL);
-	while (me) {
-		struct mapent *valid;
-		char *key = NULL, *next_key = NULL;
-
-		if (me->age >= age) {
-			/*
-			 * Reset time of last fail for valid map entries to
-			 * force entry update and subsequent mount retry.
-			 * A map entry that's still invalid after a read
-			 * may have been created by a failed wildcard lookup
-			 * so reset the status on those too.
-			 */
-			if (me->mapent || cache_lookup(mc, "*"))
-				me->status = 0;
-			me = cache_enumerate(mc, me);
-			continue;
-		}
-
-		key = strdup(me->key);
-		me = cache_enumerate(mc, me);
-		/* Don't consider any entries with a wildcard */
-		if (!key || strchr(key, '*')) {
-			if (key)
-				free(key);
-			continue;
-		}
-
-		path = make_fullpath(ap, key);
-		if (!path) {
-			warn(ap->logopt, "can't malloc storage for path");
-			free(key);
-			continue;
-		}
-
-		/*
-		 * If this key has another valid entry we want to prune it,
-		 * even if it's a mount, as the valid entry will take the
-		 * mount if it is a direct mount or it's just a stale indirect
-		 * cache entry.
-		 */
-		valid = lookup_source_valid_mapent(ap, key, LKP_DISTINCT);
-		if (valid && valid->mc == mc) {
-			 /*
-			  * We've found a map entry that has been removed from
-			  * the current cache so it isn't really valid. Set the
-			  * mapent negative to prevent further mount requests
-			  * using the cache entry.
-			  */
-			debug(ap->logopt, "removed map entry detected, mark negative");
-			if (valid->mapent) {
-				free(valid->mapent);
-				valid->mapent = NULL;
-			}
-			cache_unlock(valid->mc);
-			valid = NULL;
-		}
-		if (!valid &&
-		    is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
-			debug(ap->logopt, "prune posponed, %s mounted", path);
-			free(key);
-			free(path);
-			continue;
-		}
-		if (valid)
-			cache_unlock(valid->mc);
-
-		if (me)
-			next_key = strdup(me->key);
-
-		cache_unlock(mc);
-
-		cache_writelock(mc);
-		this = cache_lookup_distinct(mc, key);
-		if (!this) {
-			cache_unlock(mc);
-			goto next;
-		}
-
-		if (valid)
-			cache_delete(mc, key);
-		else if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
-			dev_t devid = ap->dev;
-			status = CHE_FAIL;
-			if (ap->type == LKP_DIRECT)
-				devid = this->dev;
-			if (this->ioctlfd == -1)
-				status = cache_delete(mc, key);
-			if (status != CHE_FAIL) {
-				if (ap->type == LKP_INDIRECT) {
-					if (ap->flags & MOUNT_FLAG_GHOST)
-						rmdir_path(ap, path, devid);
-				} else
-					rmdir_path(ap, path, devid);
-			}
-		}
-		cache_unlock(mc);
-
-next:
-		cache_readlock(mc);
-		if (next_key) {
-			me = cache_lookup_distinct(mc, next_key);
-			free(next_key);
-		}
-		free(key);
-		free(path);
-	}
-
-	return;
-}
-
 int lookup_prune_cache(struct autofs_point *ap, time_t age)
 {
 	struct master_mapent *entry = ap->entry;
 	struct map_source *map;
+	struct mapent_cache *mc;
+	struct mapent *me, *this;
+	char *path;
+	int status = CHE_FAIL;
 
 	pthread_cleanup_push(master_source_lock_cleanup, entry);
 	master_source_readlock(entry);
@@ -1494,15 +981,90 @@ int lookup_prune_cache(struct autofs_point *ap, time_t age)
 	map = entry->maps;
 	while (map) {
 		/* Is the map stale */
-		if (!map->stale && !check_stale_instances(map)) {
+		if (!map->stale) {
 			map = map->next;
 			continue;
 		}
-		pthread_cleanup_push(cache_lock_cleanup, map->mc);
-		cache_readlock(map->mc);
-		lookup_prune_one_cache(ap, map->mc, age);
+		mc = map->mc;
+		pthread_cleanup_push(cache_lock_cleanup, mc);
+		cache_readlock(mc);
+		me = cache_enumerate(mc, NULL);
+		while (me) {
+			char *key = NULL, *next_key = NULL;
+
+			if (me->age >= age) {
+				me = cache_enumerate(mc, me);
+				continue;
+			}
+
+			key = strdup(me->key);
+			me = cache_enumerate(mc, me);
+			if (!key || *key == '*') {
+				if (key)
+					free(key);
+				continue;
+			}
+
+			path = make_fullpath(ap->path, key);
+			if (!path) {
+				warn(ap->logopt,
+				     "can't malloc storage for path");
+				free(key);
+				continue;
+			}
+
+			if (is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
+				debug(ap->logopt,
+				      "prune check posponed, %s mounted", path);
+				free(key);
+				free(path);
+				continue;
+			}
+
+			if (me)
+				next_key = strdup(me->key);
+
+			cache_unlock(mc);
+
+			cache_writelock(mc);
+			this = cache_lookup_distinct(mc, key);
+			if (!this) {
+				cache_unlock(mc);
+				free(key);
+				if (next_key)
+					free(next_key);
+				free(path);
+				goto next;
+			}
+
+			if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
+				status = CHE_FAIL;
+				if (this->ioctlfd == -1)
+					status = cache_delete(mc, key);
+				if (status != CHE_FAIL) {
+					if (ap->type == LKP_INDIRECT) {
+						if (ap->flags & MOUNT_FLAG_GHOST)
+							rmdir_path(ap, path, ap->dev);
+					} else
+						rmdir_path(ap, path, this->dev);
+				}
+			}
+			cache_unlock(mc);
+
+			if (!next_key) {
+				free(key);
+				free(path);
+				cache_readlock(mc);
+				continue;
+			}
+next:
+			cache_readlock(mc);
+			me = cache_lookup_distinct(mc, next_key);
+			free(key);
+			free(path);
+			free(next_key);
+		}
 		pthread_cleanup_pop(1);
-		clear_stale_instances(map);
 		map->stale = 0;
 		map = map->next;
 	}
@@ -1520,6 +1082,7 @@ struct mapent *lookup_source_valid_mapent(struct autofs_point *ap, const char *k
 	struct mapent_cache *mc;
 	struct mapent *me = NULL;
 
+	master_source_readlock(entry);
 	map = entry->maps;
 	while (map) {
 		/*
@@ -1542,6 +1105,7 @@ struct mapent *lookup_source_valid_mapent(struct autofs_point *ap, const char *k
 		cache_unlock(mc);
 		map = map->next;
 	}
+	master_source_unlock(entry);
 
 	return me;
 }
