@@ -16,8 +16,7 @@
 
 #include "automount.h"
 
-/* Attribute to create detached thread */
-extern pthread_attr_t th_attr_detached;
+extern pthread_attr_t thread_attr;
 
 struct state_queue {
 	pthread_t thid;
@@ -144,7 +143,7 @@ void expire_cleanup(void *arg)
 					ap->submount = 2;
 			}
 
-			if (ap->state == ST_EXPIRE && !ap->submount)
+			if (!ap->submount)
 				alarm_add(ap, ap->exp_runfreq);
 
 			/* FALLTHROUGH */
@@ -160,7 +159,7 @@ void expire_cleanup(void *arg)
 			 * been signaled to shutdown.
 			 */
 			rv = ops->askumount(ap->logopt, ap->ioctlfd, &idle);
-			if (!rv && !idle && !ap->shutdown) {
+			if (!idle && !ap->shutdown) {
 				next = ST_READY;
 				if (!ap->submount)
 					alarm_add(ap, ap->exp_runfreq);
@@ -197,11 +196,11 @@ void expire_cleanup(void *arg)
 		}
 	}
 
-	st_set_done(ap);
-
 	if (next != ST_INVAL)
 		__st_add_task(ap, next);
 
+	st_set_done(ap);
+
 	st_mutex_unlock();
 
 	return;
@@ -293,11 +292,12 @@ static enum expire expire_proc(struct autofs_point *ap, int now)
 	else
 		expire = expire_proc_direct;
 
-	status = pthread_create(&thid, &th_attr_detached, expire, ea);
+	status = pthread_create(&thid, &thread_attr, expire, ea);
 	if (status) {
 		error(ap->logopt,
 		      "expire thread create for %s failed", ap->path);
 		expire_proc_cleanup((void *) ea);
+		free(ea);
 		return EXP_ERROR;
 	}
 	ap->exp_thread = thid;
@@ -330,9 +330,14 @@ static void do_readmap_cleanup(void *arg)
 	ap = ra->ap;
 
 	st_mutex_lock();
+
 	ap->readmap_thread = 0;
-	st_set_done(ap);
 	st_ready(ap);
+	st_set_done(ap);
+
+	if (!ap->submount)
+		alarm_add(ap, ap->exp_runfreq);
+
 	st_mutex_unlock();
 
 	free(ra);
@@ -347,96 +352,6 @@ static void tree_mnts_cleanup(void *arg)
 	return;
 }
 
-static void do_readmap_mount(struct autofs_point *ap, struct mnt_list *mnts,
-			     struct map_source *map, struct mapent *me, time_t now)
-{
-	struct mapent_cache *nc;
-	struct mapent *ne, *nested, *valid;
-
-	nc = ap->entry->master->nc;
-
-	ne = cache_lookup_distinct(nc, me->key);
-	if (!ne) {
-		nested = cache_partial_match(nc, me->key);
-		if (nested) {
-			error(ap->logopt,
-			      "removing invalid nested null entry %s",
-			      nested->key);
-			nested = cache_partial_match(nc, me->key);
-			if (nested)
-				cache_delete(nc, nested->key);
-		}
-	}
-
-	if (me->age < now || (ne && map->master_line > ne->age)) {
-		/*
-		 * The map instance may have changed, such as the map name or
-		 * the mount options, but the direct map entry may still exist
-		 * in one of the other maps. If so then update the new cache
-		 * entry device and inode so we can find it at lookup. Later,
-		 * the mount for the new cache entry will just update the
-		 * timeout.
-		 *
-		 * TODO: how do we recognise these orphaned map instances. We
-		 * can't just delete these instances when the cache becomes
-		 * empty because that is a valid state for a master map entry.
-		 * This is becuase of the requirement to continue running with
-		 * an empty cache awaiting a map re-load.
-		 */
-		valid = lookup_source_valid_mapent(ap, me->key, LKP_DISTINCT);
-		if (valid && valid->mc == me->mc) {
-			/*
-			 * We've found a map entry that has been removed from
-			 * the current cache so there is no need to update it.
-			 * The stale entry will be dealt with when we prune the
-			 * cache later.
-			 */
-			cache_unlock(valid->mc);
-			valid = NULL;
-		}
-		if (valid) {
-			struct mapent_cache *vmc = valid->mc;
-			struct ioctl_ops *ops = get_ioctl_ops();
-			time_t runfreq;
-
-			cache_unlock(vmc);
-			debug(ap->logopt,
-			     "updating cache entry for valid direct trigger %s",
-			     me->key);
-			cache_writelock(vmc);
-			valid = cache_lookup_distinct(vmc, me->key);
-			if (!valid) {
-				cache_unlock(vmc);
-				error(ap->logopt,
-				     "failed to find expected existing valid map entry");
-				return;
-			}
-			/* Take over the mount if there is one */
-			valid->ioctlfd = me->ioctlfd;
-			me->ioctlfd = -1;
-			/* Set device and inode number of the new mapent */
-			cache_set_ino_index(vmc, me->key, me->dev, me->ino);
-			cache_unlock(vmc);
-			/* Set timeout and calculate the expire run frequency */
-			ops->timeout(ap->logopt, valid->ioctlfd, map->exp_timeout);
-			if (map->exp_timeout) {
-				runfreq = (map->exp_timeout + CHECK_RATIO - 1) / CHECK_RATIO;
-				if (ap->exp_runfreq)
-					ap->exp_runfreq = min(ap->exp_runfreq, runfreq);
-				else
-					ap->exp_runfreq = runfreq;
-			}
-		} else if (!tree_is_mounted(mnts, me->key, MNTS_REAL))
-			do_umount_autofs_direct(ap, mnts, me);
-		else
-			debug(ap->logopt,
-			      "%s is mounted", me->key);
-	} else
-		do_mount_autofs_direct(ap, mnts, me, map->exp_timeout);
-
-	return;
-}
-
 static void *do_readmap(void *arg)
 {
 	struct autofs_point *ap;
@@ -480,22 +395,17 @@ static void *do_readmap(void *arg)
 	pthread_cleanup_pop(1);
 
 	if (ap->type == LKP_INDIRECT) {
-		struct ioctl_ops *ops = get_ioctl_ops();
-		time_t timeout = ap->entry->maps->exp_timeout;
-		ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
-		ops->timeout(ap->logopt, ap->ioctlfd, timeout);
 		lookup_prune_cache(ap, now);
 		status = lookup_ghost(ap, ap->path);
 	} else {
-		struct mapent *me;
-
+		struct mapent *me, *ne, *nested;
 		mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
 		pthread_cleanup_push(tree_mnts_cleanup, mnts);
+		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
+		master_source_readlock(ap->entry);
 		nc = ap->entry->master->nc;
 		cache_readlock(nc);
 		pthread_cleanup_push(cache_lock_cleanup, nc);
-		master_source_readlock(ap->entry);
-		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 		map = ap->entry->maps;
 		while (map) {
 			/* Is map source up to date or no longer valid */
@@ -508,10 +418,31 @@ static void *do_readmap(void *arg)
 			cache_readlock(mc);
 			me = cache_enumerate(mc, NULL);
 			while (me) {
-				do_readmap_mount(ap, mnts, map, me, now);
+				ne = cache_lookup_distinct(nc, me->key);
+				if (!ne) {
+					nested = cache_partial_match(nc, me->key);
+					if (nested) {
+						error(ap->logopt,
+						"removing invalid nested null entry %s",
+						nested->key);
+						nested = cache_partial_match(nc, me->key);
+						if (nested)
+							cache_delete(nc, nested->key);
+					}
+				}
+
+				/* TODO: check return of do_... */
+				if (me->age < now || (ne && map->master_line > ne->age)) {
+					if (!tree_is_mounted(mnts, me->key, MNTS_REAL))
+						do_umount_autofs_direct(ap, mnts, me);
+					else
+                                		debug(ap->logopt,
+						      "%s is mounted", me->key);
+				} else
+					do_mount_autofs_direct(ap, mnts, me);
+
 				me = cache_enumerate(mc, me);
 			}
-			lookup_prune_one_cache(ap, map->mc, now);
 			pthread_cleanup_pop(1);
 			map->stale = 0;
 			map = map->next;
@@ -519,6 +450,7 @@ static void *do_readmap(void *arg)
 		pthread_cleanup_pop(1);
 		pthread_cleanup_pop(1);
 		pthread_cleanup_pop(1);
+		lookup_prune_cache(ap, now);
 	}
 
 	pthread_cleanup_pop(1);
@@ -587,7 +519,7 @@ static unsigned int st_readmap(struct autofs_point *ap)
 	ra->ap = ap;
 	ra->now = now;
 
-	status = pthread_create(&thid, &th_attr_detached, do_readmap, ra);
+	status = pthread_create(&thid, &thread_attr, do_readmap, ra);
 	if (status) {
 		error(ap->logopt, "read map thread create failed");
 		st_readmap_cleanup(ra);
@@ -820,6 +752,7 @@ done:
 			list_add(&new->list, head);
 	}
 
+	/* Added task, encourage state machine */
 	signaled = 1;
 	status = pthread_cond_signal(&cond);
 	if (status)
@@ -850,14 +783,10 @@ void st_remove_tasks(struct autofs_point *ap)
 	struct state_queue *task, *waiting;
 	int status;
 
-	st_mutex_lock();
-
 	head = &state_queue;
 
-	if (list_empty(head)) {
-		st_mutex_unlock();
+	if (list_empty(head))
 		return;
-	}
 
 	p = head->next;
 	while (p != head) {
@@ -894,8 +823,6 @@ void st_remove_tasks(struct autofs_point *ap)
 	if (status)
 		fatal(status);
 
-	st_mutex_unlock();
-
 	return;
 }
 
@@ -1075,6 +1002,8 @@ static void *st_queue_handler(void *arg)
 {
 	struct list_head *head;
 	struct list_head *p;
+	struct timespec wait;
+	struct timeval now;
 	int status, ret;
 
 	st_mutex_lock();
@@ -1085,11 +1014,17 @@ static void *st_queue_handler(void *arg)
 		 * entry is added.
 		 */
 		head = &state_queue;
+		gettimeofday(&now, NULL);
+		wait.tv_sec = now.tv_sec + 1;
+		wait.tv_nsec = now.tv_usec * 1000;
 
 		while (list_empty(head)) {
-			status = pthread_cond_wait(&cond, &mutex);
-			if (status)
+			status = pthread_cond_timedwait(&cond, &mutex, &wait);
+			if (status) {
+				if (status == ETIMEDOUT)
+					break;
 				fatal(status);
+			}
 		}
 
 		p = head->next;
@@ -1115,11 +1050,18 @@ static void *st_queue_handler(void *arg)
 		}
 
 		while (1) {
+			gettimeofday(&now, NULL);
+			wait.tv_sec = now.tv_sec + 1;
+			wait.tv_nsec = now.tv_usec * 1000;
+
 			signaled = 0;
 			while (!signaled) {
-				status = pthread_cond_wait(&cond, &mutex);
-				if (status)
+				status = pthread_cond_timedwait(&cond, &mutex, &wait);
+				if (status) {
+					if (status == ETIMEDOUT)
+						break;
 					fatal(status);
+				}
 			}
 
 			head = &state_queue;
@@ -1130,10 +1072,6 @@ static void *st_queue_handler(void *arg)
 				task = list_entry(p, struct state_queue, list);
 				p = p->next;
 
-				/* Task may have been canceled before it started */
-				if (!task->thid && task->cancel)
-					goto remove;
-
 				if (!task->busy) {
 					/* Start a new task */
 					task->busy = 1;
@@ -1170,13 +1108,11 @@ remove:
 				next = list_entry((&task->pending)->next,
 							struct state_queue, pending);
 
+				list_del_init(&next->pending);
+				list_add_tail(&next->list, p);
+
 				list_del(&task->list);
 				free(task);
-
-				list_del_init(&next->pending);
-				list_add_tail(&next->list, head);
-				if (p == head)
-					p = head->next;
 			}
 
 			if (list_empty(head))
@@ -1203,10 +1139,9 @@ int st_start_handler(void)
 	}
 
 	status = pthread_create(&thid, pattrs, st_queue_handler, NULL);
+	if (status)
+		return 0;
 
-	if (pattrs)
-		pthread_attr_destroy(pattrs);
-
-	return !status;
+	return 1;
 }
 
