@@ -17,9 +17,12 @@
 #include <malloc.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 #include <ctype.h>
 #include <stdio.h>
+#include <fcntl.h>
 #include <sys/param.h>
+#include <sys/types.h>
 #include <sys/stat.h>
 
 #include "automount.h"
@@ -176,69 +179,6 @@ static inline void ino_index_unlock(struct mapent_cache *mc)
 	return;
 }
 
-/* Save the cache entry mapent field onto a stack and set a new mapent */
-int cache_push_mapent(struct mapent *me, char *mapent)
-{
-	struct stack *s;
-	char *new;
-
-	if (!me->mapent)
-		return CHE_FAIL;
-
-	if (!mapent)
-		new = NULL;
-	else {
-		new = strdup(mapent);
-		if (!new)
-			return CHE_FAIL;
-	}
-
-	s = malloc(sizeof(struct stack));
-	if (!s) {
-		if (new)
-			free(new);
-		return CHE_FAIL;
-	}
-	memset(s, 0, sizeof(*s));
-
-	s->mapent = me->mapent;
-	s->age = me->age;
-	me->mapent = new;
-
-	if (me->stack)
-		s->next = me->stack;
-	me->stack = s;
-
-	return CHE_OK;
-}
-
-/* Restore cache entry mapent to a previously saved mapent, discard current */
-int cache_pop_mapent(struct mapent *me)
-{
-	struct stack *s = me->stack;
-	char *mapent;
-	time_t age;
-
-	if (!s || !s->mapent)
-		return CHE_FAIL;
-
-	mapent = s->mapent;
-	age = s->age;
-	me->stack = s->next;
-	free(s);
-
-	if (age < me->age) {
-		free(mapent);
-		return CHE_OK;
-	}
-
-	if (me->mapent)
-		free(me->mapent);
-	me->mapent = mapent;
-
-	return CHE_OK;
-}
-
 struct mapent_cache *cache_init(struct autofs_point *ap, struct map_source *map)
 {
 	struct mapent_cache *mc;
@@ -252,9 +192,9 @@ struct mapent_cache *cache_init(struct autofs_point *ap, struct map_source *map)
 	if (!mc)
 		return NULL;
 
-	mc->size = defaults_get_map_hash_table_size();
+	mc->size = HASHSIZE;
 
-	mc->hash = malloc(mc->size * sizeof(struct mapent *));
+	mc->hash = malloc(mc->size * sizeof(struct entry *));
 	if (!mc->hash) {
 		free(mc);
 		return NULL;
@@ -290,46 +230,22 @@ struct mapent_cache *cache_init(struct autofs_point *ap, struct map_source *map)
 	return mc;
 }
 
-void cache_clean_null_cache(struct mapent_cache *mc)
-{
-	struct mapent *me, *next;
-	int i;
-
-	for (i = 0; i < mc->size; i++) {
-		me = mc->hash[i];
-		if (me == NULL)
-			continue;
-		next = me->next;
-		free(me->key);
-		if (me->mapent)
-			free(me->mapent);
-		free(me);
-
-		while (next != NULL) {
-			me = next;
-			next = me->next;
-			free(me->key);
-			free(me);
-		}
-		mc->hash[i] = NULL;
-	}
-
-	return;
-}
-
 struct mapent_cache *cache_init_null_cache(struct master *master)
 {
 	struct mapent_cache *mc;
 	unsigned int i;
 	int status;
 
+	if (master->nc)
+		cache_release_null_cache(master);
+
 	mc = malloc(sizeof(struct mapent_cache));
 	if (!mc)
 		return NULL;
 
-	mc->size = NULL_MAP_HASHSIZE;
+	mc->size = HASHSIZE;
 
-	mc->hash = malloc(mc->size * sizeof(struct mapent *));
+	mc->hash = malloc(mc->size * sizeof(struct entry *));
 	if (!mc->hash) {
 		free(mc);
 		return NULL;
@@ -350,6 +266,8 @@ struct mapent_cache *cache_init_null_cache(struct master *master)
 	if (status)
 		fatal(status);
 
+	cache_writelock(mc);
+
 	for (i = 0; i < mc->size; i++) {
 		mc->hash[i] = NULL;
 		INIT_LIST_HEAD(&mc->ino_index[i]);
@@ -358,21 +276,34 @@ struct mapent_cache *cache_init_null_cache(struct master *master)
 	mc->ap = NULL;
 	mc->map = NULL;
 
+	cache_unlock(mc);
+
 	return mc;
 }
 
-static u_int32_t ino_hash(dev_t dev, ino_t ino, unsigned int size)
+static unsigned int hash(const char *key)
 {
-	u_int32_t hashval;
+	unsigned long hashval;
+	char *s = (char *) key;
+
+	for (hashval = 0; *s != '\0';)
+		hashval += *s++;
+
+	return hashval % HASHSIZE;
+}
+
+static unsigned int ino_hash(dev_t dev, ino_t ino)
+{
+	unsigned long hashval;
 
 	hashval = dev + ino;
 
-	return hashval % size;
+	return hashval % HASHSIZE;
 }
 
 int cache_set_ino_index(struct mapent_cache *mc, const char *key, dev_t dev, ino_t ino)
 {
-	u_int32_t ino_index = ino_hash(dev, ino, mc->size);
+	unsigned int ino_index = ino_hash(dev, ino);
 	struct mapent *me;
 
 	me = cache_lookup_distinct(mc, key);
@@ -394,10 +325,10 @@ struct mapent *cache_lookup_ino(struct mapent_cache *mc, dev_t dev, ino_t ino)
 {
 	struct mapent *me = NULL;
 	struct list_head *head, *p;
-	u_int32_t ino_index;
+	unsigned int ino_index;
 
 	ino_index_lock(mc);
-	ino_index = ino_hash(dev, ino, mc->size);
+	ino_index = ino_hash(dev, ino);
 	head = &mc->ino_index[ino_index];
 
 	list_for_each(p, head) {
@@ -440,7 +371,7 @@ struct mapent *cache_lookup_first(struct mapent_cache *mc)
 struct mapent *cache_lookup_next(struct mapent_cache *mc, struct mapent *me)
 {
 	struct mapent *this;
-	u_int32_t hashval;
+	unsigned long hashval;
 	unsigned int i;
 
 	if (!me)
@@ -456,7 +387,7 @@ struct mapent *cache_lookup_next(struct mapent_cache *mc, struct mapent *me)
 		return this;
 	}
 
-	hashval = hash(me->key, mc->size) + 1;
+	hashval = hash(me->key) + 1;
 	if (hashval < mc->size) {
 		for (i = (unsigned int) hashval; i < mc->size; i++) {
 			this = mc->hash[i];
@@ -504,7 +435,7 @@ struct mapent *cache_lookup(struct mapent_cache *mc, const char *key)
 	if (!key)
 		return NULL;
 
-	for (me = mc->hash[hash(key, mc->size)]; me != NULL; me = me->next) {
+	for (me = mc->hash[hash(key)]; me != NULL; me = me->next) {
 		if (strcmp(key, me->key) == 0)
 			goto done;
 	}
@@ -517,7 +448,7 @@ struct mapent *cache_lookup(struct mapent_cache *mc, const char *key)
 			goto done;
 		}
 
-		for (me = mc->hash[hash("*", mc->size)]; me != NULL; me = me->next)
+		for (me = mc->hash[hash("*")]; me != NULL; me = me->next)
 			if (strcmp("*", me->key) == 0)
 				goto done;
 	}
@@ -533,7 +464,7 @@ struct mapent *cache_lookup_distinct(struct mapent_cache *mc, const char *key)
 	if (!key)
 		return NULL;
 
-	for (me = mc->hash[hash(key, mc->size)]; me != NULL; me = me->next) {
+	for (me = mc->hash[hash(key)]; me != NULL; me = me->next) {
 		if (strcmp(key, me->key) == 0)
 			return me;
 	}
@@ -546,29 +477,31 @@ struct mapent *cache_lookup_offset(const char *prefix, const char *offset, int s
 {
 	struct list_head *p;
 	struct mapent *this;
-	/* Keys for direct maps may be as long as a path name */
-	char o_key[PATH_MAX];
-	/* Avoid "//" at the beginning of paths */
-	const char *path_prefix = strlen(prefix) > 1 ? prefix : "";
-	size_t size;
+	int plen = strlen(prefix);
+	char *o_key;
 
 	/* root offset duplicates "/" */
-	size = snprintf(o_key, sizeof(o_key), "%s%s", path_prefix, offset);
-	if (size >= sizeof(o_key))
-		return NULL;
+	if (plen > 1) {
+		o_key = alloca(plen + strlen(offset) + 1);
+		strcpy(o_key, prefix);
+		strcat(o_key, offset);
+	} else {
+		o_key = alloca(strlen(offset) + 1);
+		strcpy(o_key, offset);
+	}
 
 	list_for_each(p, head) {
 		this = list_entry(p, struct mapent, multi_list);
 		if (!strcmp(&this->key[start], o_key))
-			return this;
+			goto done;
 	}
-	return NULL;
+	this = NULL;
+done:
+	return this;
 }
 
 /* cache must be read locked by caller */
-static struct mapent *__cache_partial_match(struct mapent_cache *mc,
-					    const char *prefix,
-					    unsigned int type)
+struct mapent *cache_partial_match(struct mapent_cache *mc, const char *prefix)
 {
 	struct mapent *me = NULL;
 	size_t len = strlen(prefix);
@@ -580,52 +513,26 @@ static struct mapent *__cache_partial_match(struct mapent_cache *mc,
 			continue;
 
 		if (len < strlen(me->key) &&
-		    (strncmp(prefix, me->key, len) == 0) &&
-		     me->key[len] == '/') {
-			if (type == LKP_NORMAL)
-				return me;
-			if (type == LKP_WILD &&
-			    me->key[len] != '\0' &&
-			    me->key[len + 1] == '*')
-				return me;
-		}
+		    (strncmp(prefix, me->key, len) == 0) && me->key[len] == '/')
+			return me;
 
 		me = me->next;
 		while (me != NULL) {
 			if (len < strlen(me->key) &&
-			    (strncmp(prefix, me->key, len) == 0 &&
-			    me->key[len] == '/')) {
-				if (type == LKP_NORMAL)
-					return me;
-				if (type == LKP_WILD &&
-				    me->key[len] != '\0' &&
-				    me->key[len + 1] == '*')
-					return me;
-			}
+			    strncmp(prefix, me->key, len) == 0 && me->key[len] == '/')
+				return me;
 			me = me->next;
 		}
 	}
 	return NULL;
 }
 
-/* cache must be read locked by caller */
-struct mapent *cache_partial_match(struct mapent_cache *mc, const char *prefix)
-{
-	return __cache_partial_match(mc, prefix, LKP_NORMAL);
-}
-
-/* cache must be read locked by caller */
-struct mapent *cache_partial_match_wild(struct mapent_cache *mc, const char *prefix)
-{
-	return __cache_partial_match(mc, prefix, LKP_WILD);
-}
-
 /* cache must be write locked by caller */
 int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age)
 {
 	struct mapent *me, *existing = NULL;
 	char *pkey, *pent;
-	u_int32_t hashval = hash(key, mc->size);
+	unsigned int hashval = hash(key);
 	int status;
 
 	me = (struct mapent *) malloc(sizeof(struct mapent));
@@ -650,8 +557,6 @@ int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, c
 	} else
 		me->mapent = NULL;
 
-	me->stack = NULL;
-
 	me->age = age;
 	me->status = 0;
 	me->mc = mc;
@@ -721,7 +626,7 @@ static void cache_add_ordered_offset(struct mapent *me, struct list_head *head)
 }
 
 /* cache must be write locked by caller */
-int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age)
+int cache_add_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age)
 {
 	unsigned logopt = mc->ap ? mc->ap->logopt : master_get_logopt();
 	struct mapent *me, *owner;
@@ -732,29 +637,10 @@ int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *k
 		return CHE_FAIL;
 
 	me = cache_lookup_distinct(mc, key);
-	if (me && me->age == age) {
-		if (me == owner || strcmp(me->key, key) == 0) {
-			char *pent;
+	if (me && me != owner)
+		return CHE_DUPLICATE;
 
-			warn(logopt,
-			     "duplcate offset detected for key %s", me->key);
-
-			pent = malloc(strlen(mapent) + 1);
-			if (!pent)
-				warn(logopt,
-				     "map entry not updated: %s", me->mapent);
-			else {
-				if (me->mapent)
-					free(me->mapent);
-				me->mapent = strcpy(pent, mapent);
-				warn(logopt,
-				     "map entry updated with: %s", mapent);
-			}
-			return CHE_DUPLICATE;
-		}
-	}
-
-	ret = cache_update(mc, owner->source, key, mapent, age);
+	ret = cache_add(mc, owner->source, key, mapent, age);
 	if (ret == CHE_FAIL) {
 		warn(logopt, "failed to add key %s to cache", key);
 		return CHE_FAIL;
@@ -771,32 +657,6 @@ done:
 	return ret; 
 }
 
-void cache_update_negative(struct mapent_cache *mc,
-			   struct map_source *ms, const char *key,
-			   time_t timeout)
-{
-	time_t now = time(NULL);
-	struct mapent *me;
-	int rv = CHE_OK;
-
-	/* Don't update the wildcard */
-	if (strlen(key) == 1 && *key == '*')
-		return;
-
-	me = cache_lookup_distinct(mc, key);
-	if (me)
-		rv = cache_push_mapent(me, NULL);
-	else
-		rv = cache_update(mc, ms, key, NULL, now);
-	if (rv != CHE_FAIL) {
-		me = cache_lookup_distinct(mc, key);
-		if (me)
-			me->status = now + timeout;
-	}
-	return;
-}
-
-
 static struct mapent *get_parent(const char *key, struct list_head *head, struct list_head **pos)
 {
 	struct list_head *next;
@@ -858,7 +718,7 @@ int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key
 	me = cache_lookup(mc, key);
 	while (me && me->source != ms)
 		me = cache_lookup_key_next(me);
-	if (!me || (!strcmp(me->key, "*") && strcmp(key, "*"))) {
+	if (!me || (*me->key == '*' && *key != '*')) {
 		ret = cache_add(mc, ms, key, mapent, age);
 		if (!ret) {
 			debug(logopt, "failed for %s", key);
@@ -888,61 +748,19 @@ int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key
 	return ret;
 }
 
-/* cache_multi_lock of the multi mount owner must be held by caller */
-int cache_delete_offset(struct mapent_cache *mc, const char *key)
-{
-	u_int32_t hashval = hash(key, mc->size);
-	struct mapent *me = NULL, *pred;
-	int status;
-
-	me = mc->hash[hashval];
-	if (!me)
-		return CHE_FAIL;
-
-	if (strcmp(key, me->key) == 0) {
-		if (me->multi && me->multi == me)
-			return CHE_FAIL;
-		mc->hash[hashval] = me->next;
-		goto delete;
-	}
-
-	while (me->next != NULL) {
-		pred = me;
-		me = me->next;
-		if (strcmp(key, me->key) == 0) {
-			if (me->multi && me->multi == me)
-				return CHE_FAIL;
-			pred->next = me->next;
-			goto delete;
-		}
-	}
-
-	return CHE_FAIL;
-
-delete:
-	status = pthread_rwlock_destroy(&me->multi_rwlock);
-	if (status)
-		fatal(status);
-	list_del(&me->multi_list);
-	ino_index_lock(mc);
-	list_del(&me->ino_index);
-	ino_index_unlock(mc);
-	free(me->key);
-	if (me->mapent)
-		free(me->mapent);
-	free(me);
-
-	return CHE_OK;
-}
-
 /* cache must be write locked by caller */
 int cache_delete(struct mapent_cache *mc, const char *key)
 {
 	struct mapent *me = NULL, *pred;
-	u_int32_t hashval = hash(key, mc->size);
+	unsigned int hashval = hash(key);
 	int status, ret = CHE_OK;
-	char this[PATH_MAX];
+	char *this;
 
+	this = alloca(strlen(key) + 1);
+	if (!this) {
+		ret = CHE_FAIL;
+		goto done;
+	}
 	strcpy(this, key);
 
 	me = mc->hash[hashval];
@@ -955,7 +773,6 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		pred = me;
 		me = me->next;
 		if (strcmp(this, me->key) == 0) {
-			struct stack *s = me->stack;
 			if (me->multi && !list_empty(&me->multi_list)) {
 				ret = CHE_FAIL;
 				goto done;
@@ -970,13 +787,6 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 			free(me->key);
 			if (me->mapent)
 				free(me->mapent);
-			while (s) {
-				struct stack *next = s->next;
-				if (s->mapent)
-					free(s->mapent);
-				free(s);
-				s = next;
-			}
 			free(me);
 			me = pred;
 		}
@@ -987,7 +797,6 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		goto done;
 
 	if (strcmp(this, me->key) == 0) {
-		struct stack *s = me->stack;
 		if (me->multi && !list_empty(&me->multi_list)) {
 			ret = CHE_FAIL;
 			goto done;
@@ -1002,13 +811,6 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		free(me->key);
 		if (me->mapent)
 			free(me->mapent);
-		while (s) {
-			struct stack *next = s->next;
-			if (s->mapent)
-				free(s->mapent);
-			free(s);
-			s = next;
-		}
 		free(me);
 	}
 done:
@@ -1183,6 +985,7 @@ struct mapent *cache_enumerate(struct mapent_cache *mc, struct mapent *me)
  * Get each offset from list head under prefix.
  * Maintain traversal current position in pos for subsequent calls. 
  * Return each offset into offset.
+ * TODO: length check on offset.
  */
 /* cache must be read locked by caller */
 char *cache_get_offset(const char *prefix, char *offset, int start,
@@ -1211,9 +1014,6 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			continue;
 
 		if (!strncmp(prefix, offset_start, plen)) {
-			struct mapent *np = NULL;
-			char pe[PATH_MAX + 1];
-
 			/* "/" doesn't count for root offset */
 			if (plen == 1)
 				pstart = &offset_start[plen - 1];
@@ -1226,24 +1026,7 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 
 			/* get next offset */
 			pend = pstart;
-			while (*pend++) {
-				size_t nest_pt_offset;
-
-				if (*pend != '/')
-					continue;
-
-				nest_pt_offset = start + pend - pstart;
-				if (plen > 1)
-					nest_pt_offset += plen;
-				strcpy(pe, this->key);
-				pe[nest_pt_offset] = '\0';
-
-				np = cache_lookup_distinct(this->mc, pe);
-				if (np)
-					break;
-			}
-			if (np)
-				continue;
+			while (*pend++) ;
 			len = pend - pstart - 1;
 			strncpy(offset, pstart, len);
 			offset[len] ='\0';
@@ -1276,9 +1059,7 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			break;
 
 		/* compare offset */
-		if (pstart[len] != '/' ||
-		    strlen(pstart) != len ||
-		    strncmp(offset, pstart, len))
+		if (pstart[len] != '/' || strncmp(offset, pstart, len))
 			break;
 
 		*pos = next;
