@@ -17,6 +17,8 @@
 #include <malloc.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
+#include <fcntl.h>
 #include <time.h>
 #include <ctype.h>
 #include <signal.h>
@@ -41,78 +43,20 @@ typedef enum {
 typedef enum { got_nothing, got_star, got_real, got_plus } FOUND_STATE;
 typedef enum { esc_none, esc_char, esc_val, esc_all } ESCAPES;
 
+
 struct lookup_context {
 	const char *mapname;
-	int opts_argc;
-	const char **opts_argv;
-	time_t last_read;
+	time_t mtime;
 	struct parse_mod *parse;
 };
 
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
 
-static int do_init(const char *mapfmt,
-		   int argc, const char *const *argv,
-		   struct lookup_context *ctxt, unsigned int reinit)
-{
-	int ret = 0;
-
-	if (argc < 1) {
-		logerr(MODPREFIX "No map name");
-		return 1;
-	}
-
-	ctxt->mapname = argv[0];
-
-	if (ctxt->mapname[0] != '/') {
-		logmsg(MODPREFIX
-		     "file map %s is not an absolute pathname", argv[0]);
-		return 1;
-	}
-
-	if (access(ctxt->mapname, R_OK)) {
-		warn(LOGOPT_NONE, MODPREFIX
-		    "file map %s missing or not readable", argv[0]);
-		return 1;
-	}
-
-	if (!mapfmt)
-		mapfmt = MAPFMT_DEFAULT;
-
-	argc--;
-	argv++;
-
-	ctxt->opts_argv = copy_argv(argc, (const char **) argv);
-	if (ctxt->opts_argv == NULL) {
-		warn(LOGOPT_NONE, MODPREFIX "failed to duplicate options");
-		return 1;
-	}
-	ctxt->opts_argc = argc;
-
-	if (reinit) {
-		ret = reinit_parse(ctxt->parse, mapfmt, MODPREFIX, argc, argv);
-		if (ret)
-			logmsg(MODPREFIX "failed to reinit parse context");
-	} else {
-		ctxt->parse = open_parse(mapfmt, MODPREFIX, argc, argv);
-		if (!ctxt->parse) {
-			logmsg(MODPREFIX "failed to open parse context");
-			ret = 1;
-		}
-	}
-
-	if (ret)
-		free_argv(ctxt->opts_argc, ctxt->opts_argv);
-
-	return ret;
-}
-
-int lookup_init(const char *mapfmt,
-		int argc, const char *const *argv,
-		void **context)
+int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
 {
 	struct lookup_context *ctxt;
 	char buf[MAX_ERR_BUF];
+	struct stat st;
 
 	*context = NULL;
 
@@ -122,49 +66,52 @@ int lookup_init(const char *mapfmt,
 		logerr(MODPREFIX "malloc: %s", estr);
 		return 1;
 	}
-	memset(ctxt, 0, sizeof(struct lookup_context));
 
-	if (do_init(mapfmt, argc, argv, ctxt, 0)) {
+	if (argc < 1) {
 		free(ctxt);
+		logerr(MODPREFIX "No map name");
 		return 1;
 	}
 
+	ctxt->mapname = argv[0];
+
+	if (ctxt->mapname[0] != '/') {
+		free(ctxt);
+		logmsg(MODPREFIX
+		     "file map %s is not an absolute pathname", argv[0]);
+		return 1;
+	}
+
+	if (access(ctxt->mapname, R_OK)) {
+		free(ctxt);
+		warn(LOGOPT_NONE, MODPREFIX
+		    "file map %s missing or not readable", argv[0]);
+		return 1;
+	}
+
+	if (stat(ctxt->mapname, &st)) {
+		free(ctxt);
+		logmsg(MODPREFIX "file map %s, could not stat",
+		     argv[0]);
+		return 1;
+	}
+		
+	ctxt->mtime = st.st_mtime;
+
+	if (!mapfmt)
+		mapfmt = MAPFMT_DEFAULT;
+
+	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
+	if (!ctxt->parse) {
+		free(ctxt);
+		logmsg(MODPREFIX "failed to open parse context");
+		return 1;
+	}
 	*context = ctxt;
 
 	return 0;
 }
 
-int lookup_reinit(const char *mapfmt,
-		  int argc, const char *const *argv, void **context)
-{
-	struct lookup_context *ctxt = (struct lookup_context *) *context;
-	struct lookup_context *new;
-	char buf[MAX_ERR_BUF];
-	int ret;
-
-	new = malloc(sizeof(struct lookup_context));
-	if (!new) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-	memset(new, 0, sizeof(struct lookup_context));
-
-	new->parse = ctxt->parse;
-	ret = do_init(mapfmt, argc, argv, new, 1);
-	if (ret) {
-		free(new);
-		return 1;
-	}
-
-	*context = new;
-
-	free_argv(ctxt->opts_argc, ctxt->opts_argv);
-	free(ctxt);
-
-	return 0;
-}
-
 static int read_one(unsigned logopt, FILE *f, char *key, unsigned int *k_len, char *mapent, unsigned int *m_len)
 {
 	char *kptr, *p;
@@ -355,10 +302,8 @@ static int read_one(unsigned logopt, FILE *f, char *key, unsigned int *k_len, ch
 				if (gotten == got_real || gotten == getting)
 					goto got_it;
 			} else if (mapent_len < MAPENT_MAX_LEN) {
-				if (p) {
-					mapent_len++;
-					*(p++) = ch;
-				}
+				mapent_len++;
+				*(p++) = ch;
 				nch = getc(f);
 				if (nch == EOF &&
 				   (gotten == got_real || gotten == getting))
@@ -446,15 +391,16 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	unsigned int logopt = master->logopt;
 	char *buffer;
 	int blen;
-	char path[KEY_MAX_LEN + 1];
-	char ent[MAPENT_MAX_LEN + 1];
+	char *path;
+	char *ent;
+	struct stat st;
 	FILE *f;
+	int fd, cl_flags;
 	unsigned int path_len, ent_len;
 	int entry, cur_state;
 
-	/* Don't return fail on self include, skip source */
 	if (master->recurse)
-		return NSS_STATUS_TRYAGAIN;
+		return NSS_STATUS_UNAVAIL;
 
 	if (master->depth > MAX_INCLUDE_DEPTH) {
 		error(logopt, MODPREFIX
@@ -462,16 +408,35 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 		return NSS_STATUS_UNAVAIL;
 	}
 
-	f = open_fopen_r(ctxt->mapname);
+	path = alloca(KEY_MAX_LEN + 1);
+	if (!path) {
+		error(logopt,
+		      MODPREFIX "could not malloc storage for path");
+		return NSS_STATUS_UNAVAIL;
+	}
+
+	ent = alloca(MAPENT_MAX_LEN + 1);
+	if (!ent) {
+		error(logopt,
+		      MODPREFIX "could not malloc storage for mapent");
+		return NSS_STATUS_UNAVAIL;
+	}
+
+	f = fopen(ctxt->mapname, "r");
 	if (!f) {
-		if (errno == ENOENT)
-			return NSS_STATUS_NOTFOUND;
 		error(logopt,
 		      MODPREFIX "could not open master map file %s",
 		      ctxt->mapname);
 		return NSS_STATUS_UNAVAIL;
 	}
 
+	fd = fileno(f);
+
+	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(fd, F_SETFD, cl_flags);
+	}
+
 	while(1) {
 		entry = read_one(logopt, f, path, &path_len, ent, &ent_len);
 		if (!entry) {
@@ -501,21 +466,20 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 
 			inc = check_master_self_include(master, ctxt);
 			if (inc) 
-				master->recurse = 1;
+				master->recurse = 1;;
 			master->depth++;
 			status = lookup_nss_read_master(master, age);
-			if (status != NSS_STATUS_SUCCESS) {
+			if (!status) {
 				warn(logopt,
 				     MODPREFIX
 				     "failed to read included master map %s",
 				     master->name);
-				if (status != NSS_STATUS_NOTFOUND) {
-					/*
-					 * If we're starting up we need the whole
-					 * master map initially, so tell the upper
-					 * layer to retry.
-					 */
-					master->read_fail = 1;
+				if (!master->recurse) {
+					master->name = save_name;
+					master->depth--;
+					master->recurse = 0;
+					fclose(f);
+					return NSS_STATUS_UNAVAIL;
 				}
 			}
 			master->depth--;
@@ -547,6 +511,13 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 			break;
 	}
 
+	if (fstat(fd, &st)) {
+		crit(logopt, MODPREFIX "file map %s, could not stat",
+		       ctxt->mapname);
+		return NSS_STATUS_UNAVAIL;
+	}
+	ctxt->mtime = st.st_mtime;
+
 	fclose(f);
 
 	return NSS_STATUS_SUCCESS;
@@ -591,17 +562,18 @@ static int check_self_include(const char *key, struct lookup_context *ctxt)
 }
 
 static struct map_source *
-prepare_plus_include(struct autofs_point *ap,
-		     struct map_source *source,
-		     time_t age, char *key, unsigned int inc,
-		     struct lookup_context *ctxt)
+prepare_plus_include(struct autofs_point *ap, time_t age, char *key, unsigned int inc)
 {
-	struct map_source *new;
-	struct map_type_info *info;
+	struct map_source *current;
+	struct map_source *source;
+	char *type, *map, *fmt;
 	const char *argv[2];
-	char **tmp_argv, **tmp_opts;
 	int argc;
-	char *buf;
+	char *buf, *tmp;
+
+	current = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
 
 	/*
 	 * TODO:
@@ -619,74 +591,55 @@ prepare_plus_include(struct autofs_point *ap,
 		return NULL;
 	}
 
-	if (!(info = parse_map_type_info(buf))) {
-		error(ap->logopt, MODPREFIX "failed to parse map info");
-		free(buf);
-		return NULL;
+	type = fmt = NULL;
+
+	/* Look for space terminator - ignore local options */
+	map = buf;
+	for (tmp = buf; *tmp; tmp++) {
+		if (*tmp == ' ') {
+			*tmp = '\0';
+			break;
+		} else if (*tmp == ',') {
+			type = buf;
+			*tmp++ = '\0';
+			fmt = tmp;
+		} else if (*tmp == ':') {
+			if (!fmt)
+				type = buf;
+			*tmp++ = '\0';
+			map = tmp;
+		}
+		if (*tmp == '\\')
+			tmp++;
 	}
 
 	argc = 1;
-	argv[0] = info->map;
+	argv[0] = map;
 	argv[1] = NULL;
 
-	tmp_argv = (char **) copy_argv(argc, argv);
-	if (!tmp_argv) {
-		error(ap->logopt, MODPREFIX "failed to allocate args vector");
-		free_map_type_info(info);
-		free(buf);
-		return NULL;
-	}
-
-	tmp_opts = (char **) copy_argv(ctxt->opts_argc, ctxt->opts_argv);
-	if (!tmp_opts) {
-		error(ap->logopt, MODPREFIX "failed to allocate options args vector");
-		free_argv(argc, (const char **) tmp_argv);
-		free_map_type_info(info);
-		free(buf);
-		return NULL;
-	}
-
-	tmp_argv = append_argv(argc, tmp_argv, ctxt->opts_argc, tmp_opts);
-	if (!tmp_argv) {
-		error(ap->logopt, MODPREFIX "failed to append options vector");
-		free_map_type_info(info);
-		free(buf);
-		return NULL;
-	}
-	argc += ctxt->opts_argc;
-
-	new = master_find_source_instance(source,
-					  info->type, info->format,
-					  argc, (const char **) tmp_argv);
-	if (new) {
+	source = master_find_source_instance(current, type, fmt, argc, argv);
+	if (source)
 		/*
 		 * Make sure included map age is in sync with its owner
 		 * or we could incorrectly wipe out its entries.
 		 */
-		new->age = age;
-		new->stale = 1;
-	} else {
-		new = master_add_source_instance(source,
-						 info->type, info->format, age,
-						 argc, (const char **) tmp_argv);
-		if (!new) {
-			free_argv(argc, (const char **) tmp_argv);
-			free_map_type_info(info);
+		source->age = age;
+	else {
+		source = master_add_source_instance(current, type, fmt, age, argc, argv);
+		if (!source) {
 			free(buf);
 			error(ap->logopt, "failed to add included map instance");
 			return NULL;
 		}
 	}
-	free_argv(argc, (const char **) tmp_argv);
 
-	new->depth = source->depth + 1;
+	source->depth = current->depth + 1;
 	if (inc)
-		new->recurse = 1;
+		source->recurse = 1;
 
-	free_map_type_info(info);
 	free(buf);
 
-	return new;
+	return source;
 }
 
 int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
@@ -694,9 +647,11 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	struct lookup_context *ctxt = (struct lookup_context *) context;
 	struct map_source *source;
 	struct mapent_cache *mc;
-	char key[KEY_MAX_LEN + 1];
-	char mapent[MAPENT_MAX_LEN + 1];
+	char *key;
+	char *mapent;
+	struct stat st;
 	FILE *f;
+	int fd, cl_flags;
 	unsigned int k_len, m_len;
 	int entry;
 
@@ -707,7 +662,7 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	mc = source->mc;
 
 	if (source->recurse)
-		return NSS_STATUS_TRYAGAIN;
+		return NSS_STATUS_UNAVAIL;
 
 	if (source->depth > MAX_INCLUDE_DEPTH) {
 		error(ap->logopt,
@@ -715,15 +670,34 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 		return NSS_STATUS_UNAVAIL;
 	}
 
-	f = open_fopen_r(ctxt->mapname);
+	key = alloca(KEY_MAX_LEN + 1);
+	if (!key) {
+		error(ap->logopt,
+		      MODPREFIX "could not malloc storage for key");
+		return NSS_STATUS_UNAVAIL;
+	}
+
+	mapent = alloca(MAPENT_MAX_LEN + 1);
+	if (!mapent) {
+		error(ap->logopt,
+		      MODPREFIX "could not malloc storage for mapent");
+		return NSS_STATUS_UNAVAIL;
+	}
+
+	f = fopen(ctxt->mapname, "r");
 	if (!f) {
-		if (errno == ENOENT)
-			return NSS_STATUS_NOTFOUND;
 		error(ap->logopt,
 		      MODPREFIX "could not open map file %s", ctxt->mapname);
 		return NSS_STATUS_UNAVAIL;
 	}
 
+	fd = fileno(f);
+
+	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(fd, F_SETFD, cl_flags);
+	}
+
 	while(1) {
 		entry = read_one(ap->logopt, f, key, &k_len, mapent, &m_len);
 		if (!entry) {
@@ -750,8 +724,10 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 
 			inc = check_self_include(key, ctxt);
 
-			inc_source = prepare_plus_include(ap, source,
-							  age, key, inc, ctxt);
+			master_source_current_wait(ap->entry);
+			ap->entry->current = source;
+
+			inc_source = prepare_plus_include(ap, age, key, inc);
 			if (!inc_source) {
 				debug(ap->logopt,
 				      "failed to select included map %s", key);
@@ -763,26 +739,15 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 			if (!status) {
 				warn(ap->logopt,
 				     "failed to read included map %s", key);
+				fclose(f);
+				return NSS_STATUS_UNAVAIL;
 			}
 		} else {
 			char *s_key; 
 
-			if (source->flags & MAP_FLAG_FORMAT_AMD) {
-				if (!strcmp(key, "/defaults")) {
-					cache_writelock(mc);
-					cache_update(mc, source, key, mapent, age);
-					cache_unlock(mc);
-					continue;
-				}
-				/* Don't fail on "/" in key => type == 0 */
-				s_key = sanitize_path(key, k_len, 0, ap->logopt);
-				if (!s_key)
-					continue;
-			} else {
-				s_key = sanitize_path(key, k_len, ap->type, ap->logopt);
-				if (!s_key)
-					continue;
-			}
+			s_key = sanitize_path(key, k_len, ap->type, ap->logopt);
+			if (!s_key)
+				continue;
 
 			cache_writelock(mc);
 			cache_update(mc, source, s_key, mapent, age);
@@ -795,97 +760,54 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 			break;
 	}
 
+	if (fstat(fd, &st)) {
+		crit(ap->logopt,
+		     MODPREFIX "file map %s, could not stat",
+		     ctxt->mapname);
+		return NSS_STATUS_UNAVAIL;
+	}
+	ctxt->mtime = st.st_mtime;
 	source->age = age;
-	ctxt->last_read = time(NULL);
 
 	fclose(f);
 
 	return NSS_STATUS_SUCCESS;
 }
 
-static int match_key(struct autofs_point *ap,
-		     struct map_source *source, char *map_key,
-		     const char *key, size_t key_len, const char *mapent)
-{
-	char buf[MAX_ERR_BUF];
-	struct mapent_cache *mc;
-	time_t age = monotonic_time(NULL);
-	char *lkp_key;
-	char *prefix;
-	size_t map_key_len;
-	int ret, eq;
-
-	mc = source->mc;
-
-	/* exact match is a match for both autofs and amd */
-	eq = strcmp(map_key, key);
-	if (eq == 0) {
-		cache_writelock(mc);
-		ret = cache_update(mc, source, key, mapent, age);
-		cache_unlock(mc);
-		return ret;
-	}
-
-	if (!(source->flags & MAP_FLAG_FORMAT_AMD))
-		return CHE_FAIL;
-
-	map_key_len = strlen(map_key);
-
-	lkp_key = strdup(key);
-	if (!lkp_key) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "strdup: %s", estr);
-		return CHE_FAIL;
-	}
-
-	ret = CHE_FAIL;
-
-	if (map_key_len > (strlen(lkp_key) + 2))
-		goto done;
-
-	/*
-	 * Now strip successive directory components and try a
-	 * match against map entries ending with a wildcard and
-	 * finally try the wilcard entry itself. If we get a match
-	 * then update the cache with the read key and its mapent.
-	*/
-	while ((prefix = strrchr(lkp_key, '/'))) {
-		size_t len;
-		*prefix = '\0';
-		len = strlen(lkp_key);
-		eq = strncmp(map_key, lkp_key, len);
-		if (!eq && map_key[len + 1] == '*') {
-			cache_writelock(mc);
-			ret = cache_update(mc, source, map_key, mapent, age);
-			cache_unlock(mc);
-			goto done;
-		}
-	}
-done:
-	free(lkp_key);
-	return ret;
-}
-
 static int lookup_one(struct autofs_point *ap,
-		      struct map_source *source,
 		      const char *key, int key_len,
 		      struct lookup_context *ctxt)
 {
-	struct mapent_cache *mc = source->mc;
+	struct map_source *source;
+	struct mapent_cache *mc;
 	char mkey[KEY_MAX_LEN + 1];
 	char mapent[MAPENT_MAX_LEN + 1];
-	time_t age = monotonic_time(NULL);
+	time_t age = time(NULL);
 	FILE *f;
+	int fd, cl_flags;
 	unsigned int k_len, m_len;
 	int entry, ret;
 
-	f = open_fopen_r(ctxt->mapname);
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
+	mc = source->mc;
+
+	f = fopen(ctxt->mapname, "r");
 	if (!f) {
 		error(ap->logopt,
 		      MODPREFIX "could not open map file %s", ctxt->mapname);
 		return CHE_FAIL;
 	}
 
+	fd = fileno(f);
+
+	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(fd, F_SETFD, cl_flags);
+	}
+
 	while(1) {
 		entry = read_one(ap->logopt, f, mkey, &k_len, mapent, &m_len);
 		if (entry) {
@@ -903,8 +825,10 @@ static int lookup_one(struct autofs_point *ap,
 
 				inc = check_self_include(mkey, ctxt);
 
-				inc_source = prepare_plus_include(ap, source,
-								  age, mkey, inc, ctxt);
+				master_source_current_wait(ap->entry);
+				ap->entry->current = source;
+
+				inc_source = prepare_plus_include(ap, age, mkey, inc);
 				if (!inc_source) {
 					debug(ap->logopt,
 					      MODPREFIX
@@ -921,38 +845,29 @@ static int lookup_one(struct autofs_point *ap,
 				}
 			} else {
 				char *s_key; 
+				int eq;
 
-				if (source->flags & MAP_FLAG_FORMAT_AMD) {
-					if (!strcmp(mkey, "/defaults")) {
-						cache_writelock(mc);
-						cache_update(mc, source, mkey, mapent, age);
-						cache_unlock(mc);
-						continue;
-					}
-					/* Don't fail on "/" in key => type == 0 */
-					s_key = sanitize_path(mkey, k_len, 0, ap->logopt);
-					if (!s_key)
-						continue;
-				} else {
-					s_key = sanitize_path(mkey, k_len, ap->type, ap->logopt);
-					if (!s_key)
-						continue;
+				s_key = sanitize_path(mkey, k_len, ap->type, ap->logopt);
+				if (!s_key)
+					continue;
 
-					if (key_len != strlen(s_key)) {
-						free(s_key);
-						continue;
-					}
+				if (key_len != strlen(s_key)) {
+					free(s_key);
+					continue;
 				}
 
-				ret = match_key(ap, source,
-						s_key, key, key_len, mapent);
-				if (ret == CHE_FAIL) {
+				eq = strncmp(s_key, key, key_len);
+				if (eq != 0) {
 					free(s_key);
 					continue;
 				}
 
 				free(s_key);
 
+				cache_writelock(mc);
+				ret = cache_update(mc, source, key, mapent, age);
+				cache_unlock(mc);
+
 				fclose(f);
 
 				return ret;
@@ -974,26 +889,38 @@ static int lookup_one(struct autofs_point *ap,
 	return CHE_MISSING;
 }
 
-static int lookup_wild(struct autofs_point *ap,
-		       struct map_source *source, struct lookup_context *ctxt)
+static int lookup_wild(struct autofs_point *ap, struct lookup_context *ctxt)
 {
+	struct map_source *source;
 	struct mapent_cache *mc;
 	char mkey[KEY_MAX_LEN + 1];
 	char mapent[MAPENT_MAX_LEN + 1];
-	time_t age = monotonic_time(NULL);
+	time_t age = time(NULL);
 	FILE *f;
+	int fd, cl_flags;
 	unsigned int k_len, m_len;
 	int entry, ret;
 
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
 	mc = source->mc;
 
-	f = open_fopen_r(ctxt->mapname);
+	f = fopen(ctxt->mapname, "r");
 	if (!f) {
 		error(ap->logopt,
 		      MODPREFIX "could not open map file %s", ctxt->mapname);
 		return CHE_FAIL;
 	}
 
+	fd = fileno(f);
+
+	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(fd, F_SETFD, cl_flags);
+	}
+
 	while(1) {
 		entry = read_one(ap->logopt, f, mkey, &k_len, mapent, &m_len);
 		if (entry) {
@@ -1028,55 +955,74 @@ static int lookup_wild(struct autofs_point *ap,
 }
 
 static int check_map_indirect(struct autofs_point *ap,
-			      struct map_source *source,
 			      char *key, int key_len,
 			      struct lookup_context *ctxt)
 {
+	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *exists;
 	int ret = CHE_OK;
 
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
 	mc = source->mc;
 
-	ret = lookup_one(ap, source, key, key_len, ctxt);
+	master_source_current_wait(ap->entry);
+	ap->entry->current = source;
+
+	ret = lookup_one(ap, key, key_len, ctxt);
 	if (ret == CHE_COMPLETED)
 		return NSS_STATUS_COMPLETED;
 
 	if (ret == CHE_FAIL)
 		return NSS_STATUS_NOTFOUND;
 
+	if (ret & CHE_UPDATED)
+		source->stale = 1;
+
+	pthread_cleanup_push(cache_lock_cleanup, mc);
 	cache_writelock(mc);
-	if (source->flags & MAP_FLAG_FORMAT_AMD)
-		exists = match_cached_key(ap, MODPREFIX, source, key);
-	else
-		exists = cache_lookup_distinct(mc, key);
+	exists = cache_lookup_distinct(mc, key);
 	/* Not found in the map but found in the cache */
 	if (exists && exists->source == source && ret & CHE_MISSING) {
 		if (exists->mapent) {
 			free(exists->mapent);
 			exists->mapent = NULL;
 			exists->status = 0;
+			source->stale = 1;
 		}
 	}
-	cache_unlock(mc);
+	pthread_cleanup_pop(1);
 
 	if (ret == CHE_MISSING) {
 		struct mapent *we;
 		int wild = CHE_MISSING;
 
-		wild = lookup_wild(ap, source, ctxt);
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+
+		wild = lookup_wild(ap, ctxt);
 		/*
 		 * Check for map change and update as needed for
 		 * following cache lookup.
 		 */
+		pthread_cleanup_push(cache_lock_cleanup, mc);
 		cache_writelock(mc);
 		we = cache_lookup_distinct(mc, "*");
 		if (we) {
 			/* Wildcard entry existed and is now gone */
-			if (we->source == source && (wild & CHE_MISSING))
+			if (we->source == source && (wild & CHE_MISSING)) {
 				cache_delete(mc, "*");
+				source->stale = 1;
+			}
+		} else {
+			/* Wildcard not in map but now is */
+			if (wild & (CHE_OK | CHE_UPDATED))
+				source->stale = 1;
 		}
-		cache_unlock(mc);
+		pthread_cleanup_pop(1);
 
 		if (wild & (CHE_OK | CHE_UPDATED))
 			return NSS_STATUS_SUCCESS;
@@ -1088,46 +1034,6 @@ static int check_map_indirect(struct autofs_point *ap,
 	return NSS_STATUS_SUCCESS;
 }
 
-static int map_update_needed(struct autofs_point *ap,
-			     struct map_source *source,
-			     struct lookup_context *ctxt)
-{
-	struct stat st;
-	int ret = 1;
-
-	/*
-	 * We can skip the map lookup and cache update altogether
-	 * if we know the map hasn't been modified since it was
-	 * last read. If it has then we can mark the map stale
-	 * so a re-read is triggered following the lookup.
-	 */
-	if (stat(ctxt->mapname, &st)) {
-		error(ap->logopt, MODPREFIX
-		      "file map %s, could not stat", ctxt->mapname);
-		return -1;
-	}
-
-	if (st.st_mtime <= ctxt->last_read) {
-		/*
-		 * If any map instances are present for this source
-		 * then either we have plus included entries or we
-		 * are looking through the list of nsswitch sources.
-		 * In either case, or if it's a "multi" source, we
-		 * cannot avoid reading through the map because we
-		 * must preserve the key order over multiple sources
-		 * or maps. But also, we can't know, at this point,
-		 * if a source instance has been changed since the
-		 * last time we checked it.
-		 */
-		if (!source->instance &&
-		    source->type && strcmp(source->type, "multi"))
-			ret = 0;
-	} else
-		source->stale = 1;
-
-	return ret;
-}
-
 int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
@@ -1136,10 +1042,8 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	struct mapent *me;
 	char key[KEY_MAX_LEN + 1];
 	int key_len;
-	char *lkp_key;
 	char *mapent = NULL;
-	char mapent_buf[MAPENT_MAX_LEN + 1];
-	char buf[MAX_ERR_BUF];
+	int mapent_len;
 	int status = 0;
 	int ret = 1;
 
@@ -1161,44 +1065,18 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 
 	debug(ap->logopt, MODPREFIX "looking up %s", name);
 
-	if (!(source->flags & MAP_FLAG_FORMAT_AMD)) {
-		key_len = snprintf(key, KEY_MAX_LEN + 1, "%s", name);
-		if (key_len > KEY_MAX_LEN)
-			return NSS_STATUS_NOTFOUND;
-	} else {
-		key_len = expandamdent(name, NULL, NULL);
-		if (key_len > KEY_MAX_LEN)
-			return NSS_STATUS_NOTFOUND;
-		memset(key, 0, KEY_MAX_LEN + 1);
-		expandamdent(name, key, NULL);
-		debug(ap->logopt, MODPREFIX "expanded key: \"%s\"", key);
-	}
+	key_len = snprintf(key, KEY_MAX_LEN + 1, "%s", name);
+	if (key_len > KEY_MAX_LEN)
+		return NSS_STATUS_NOTFOUND;
 
-	/* Check if we recorded a mount fail for this key anywhere */
-	me = lookup_source_mapent(ap, key, LKP_DISTINCT);
-	if (me) {
-		if (me->status >= monotonic_time(NULL)) {
-			cache_unlock(me->mc);
-			return NSS_STATUS_NOTFOUND;
-		} else {
-			struct mapent_cache *smc = me->mc;
-			struct mapent *sme;
-
-			if (me->mapent)
-				cache_unlock(smc);
-			else {
-				cache_unlock(smc);
-				cache_writelock(smc);
-				sme = cache_lookup_distinct(smc, key);
-				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent) {
-					if (cache_pop_mapent(sme) == CHE_FAIL)
-						cache_delete(smc, key);
-				}
-				cache_unlock(smc);
-			}
-		}
+	/* Check if we recorded a mount fail for this key */
+	cache_readlock(mc);
+	me = cache_lookup_distinct(mc, key);
+	if (me && me->status >= time(NULL)) {
+		cache_unlock(mc);
+		return NSS_STATUS_UNAVAIL;
 	}
+	cache_unlock(mc);
 
 	/*
 	 * We can't check the direct mount map as if it's not in
@@ -1206,38 +1084,23 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	 * we never know about it.
 	 */
 	if (ap->type == LKP_INDIRECT && *key != '/') {
-		int ret;
-
-		ret = map_update_needed(ap, source, ctxt);
-		if (!ret)
-			goto do_cache_lookup;
-		/* Map isn't accessable, just try the cache */
-		if (ret < 0)
-			goto do_cache_lookup;
+		char *lkp_key;
 
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, key);
 		if (me && me->multi)
 			lkp_key = strdup(me->multi->key);
-		else if (!ap->pref)
+		else
 			lkp_key = strdup(key);
-		else {
-			lkp_key = malloc(strlen(ap->pref) + strlen(key) + 1);
-			if (lkp_key) {
-				strcpy(lkp_key, ap->pref);
-				strcat(lkp_key, key);
-			}
-		}
 		cache_unlock(mc);
 
-		if (!lkp_key) {
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			error(ap->logopt, MODPREFIX "malloc: %s", estr);
+		if (!lkp_key)
 			return NSS_STATUS_UNKNOWN;
-		}
 
-		status = check_map_indirect(ap, source,
-					    lkp_key, strlen(lkp_key), ctxt);
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+
+		status = check_map_indirect(ap, lkp_key, strlen(lkp_key), ctxt);
 		free(lkp_key);
 		if (status) {
 			if (status == NSS_STATUS_COMPLETED)
@@ -1246,74 +1109,50 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		}
 	}
 
-	/*
-	 * We can't take the writelock for direct mounts. If we're
-	 * starting up or trying to re-connect to an existing direct
-	 * mount we could be iterating through the map entries with
-	 * the readlock held. But we don't need to update the cache
-	 * when we're starting up so just take the readlock in that
-	 * case.
-	 */
-do_cache_lookup:
-	if (ap->flags & MOUNT_FLAG_REMOUNT)
-		cache_readlock(mc);
-	else
-		cache_writelock(mc);
-
-	if (!ap->pref)
-		lkp_key = strdup(key);
-	else {
-		lkp_key = malloc(strlen(ap->pref) + strlen(key) + 1);
-		if (lkp_key) {
-			strcpy(lkp_key, ap->pref);
-			strcat(lkp_key, key);
-		}
+	cache_readlock(mc);
+	me = cache_lookup(mc, key);
+	/* Stale mapent => check for entry in alternate source or wildcard */
+	if (me && !me->mapent) {
+		while ((me = cache_lookup_key_next(me)))
+			if (me->source == source)
+				break;
+		if (!me)
+			me = cache_lookup_distinct(mc, "*");
 	}
-
-	if (!lkp_key) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "malloc: %s", estr);
-		cache_unlock(mc);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	me = match_cached_key(ap, MODPREFIX, source, lkp_key);
-	if (me && me->mapent) {
-		if (me && (me->source == source || *me->key == '/')) {
-			strcpy(mapent_buf, me->mapent);
-			mapent = mapent_buf;
-		}
+	if (me && (me->source == source || *me->key == '/')) {
+		pthread_cleanup_push(cache_lock_cleanup, mc);
+		mapent_len = strlen(me->mapent);
+		mapent = alloca(mapent_len + 1);
+		strcpy(mapent, me->mapent);
+		pthread_cleanup_pop(0);
 	}
 	cache_unlock(mc);
 
-	if (!me) {
-		free(lkp_key);
-		return NSS_STATUS_NOTFOUND;
+	if (mapent) {
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+
+		debug(ap->logopt, MODPREFIX "%s -> %s", key, mapent);
+		ret = ctxt->parse->parse_mount(ap, key, key_len,
+					mapent, ctxt->parse->context);
+		if (ret) {
+			time_t now = time(NULL);
+			int rv = CHE_OK;
+
+			cache_writelock(mc);
+			me = cache_lookup_distinct(mc, key);
+			if (!me)
+				rv = cache_update(mc, source, key, NULL, now);
+			if (rv != CHE_FAIL) {
+				me = cache_lookup_distinct(mc, key);
+				me->status = now + ap->negative_timeout;
+			}
+			cache_unlock(mc);
+		}
 	}
 
-	if (!mapent) {
-		free(lkp_key);
+	if (ret)
 		return NSS_STATUS_TRYAGAIN;
-	}
-
-	debug(ap->logopt, MODPREFIX "%s -> %s", lkp_key, mapent);
-
-	free(lkp_key);
-
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
-
-	ret = ctxt->parse->parse_mount(ap, key, key_len,
-				       mapent, ctxt->parse->context);
-	if (ret) {
-		/* Don't update negative cache when re-connecting */
-		if (ap->flags & MOUNT_FLAG_REMOUNT)
-			return NSS_STATUS_TRYAGAIN;
-		cache_writelock(mc);
-		cache_update_negative(mc, source, key, ap->negative_timeout);
-		cache_unlock(mc);
-		return NSS_STATUS_TRYAGAIN;
-	}
 
 	return NSS_STATUS_SUCCESS;
 }
@@ -1322,7 +1161,6 @@ int lookup_done(void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
 	int rv = close_parse(ctxt->parse);
-	free_argv(ctxt->opts_argc, ctxt->opts_argv);
 	free(ctxt);
 	return rv;
 }
