@@ -17,6 +17,8 @@
 #include <stdio.h>
 #include <malloc.h>
 #include <netdb.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/param.h>
@@ -58,20 +60,14 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		const char *what, const char *fstype, const char *options,
 		void *context)
 {
-	char fullpath[PATH_MAX];
-	char buf[MAX_ERR_BUF];
+	char *fullpath, buf[MAX_ERR_BUF];
 	struct host *this, *hosts = NULL;
-	unsigned int mount_default_proto, vers;
+	unsigned int vers;
 	char *nfsoptions = NULL;
-	const char *port_opt = NULL;
-	unsigned int flags = ap->flags &
-			(MOUNT_FLAG_RANDOM_SELECT | MOUNT_FLAG_USE_WEIGHT_ONLY);
-	int nobind = ap->flags & MOUNT_FLAG_NOBIND;
+	unsigned int random_selection = ap->flags & MOUNT_FLAG_RANDOM_SELECT;
 	int len, status, err, existed = 1;
 	int nosymlink = 0;
-	int port = -1;
 	int ro = 0;            /* Set if mount bind should be read-only */
-	int rdma = 0;
 
 	if (ap->flags & MOUNT_FLAG_REMOUNT)
 		return 0;
@@ -80,13 +76,6 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	      MODPREFIX "root=%s name=%s what=%s, fstype=%s, options=%s",
 	      root, name, what, fstype, options);
 
-	mount_default_proto = defaults_get_mount_nfs_default_proto();
-	vers = NFS_VERS_MASK | NFS_PROTO_MASK;
-	if (strcmp(fstype, "nfs4") == 0)
-		vers = NFS4_VERS_MASK | TCP_SUPPORTED;
-	else if (mount_default_proto == 4)
-		vers = vers | NFS4_VERS_MASK;
-
 	/* Extract "nosymlink" pseudo-option which stops local filesystems
 	 * from being symlinked.
 	 *
@@ -124,56 +113,12 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 			while (*comma == ' ' || *comma == '\t')
 				end--;
 
-			o_len = end - cp + 1;
-
-			if (_strncmp("proto=rdma", cp, o_len) == 0 ||
-				   _strncmp("rdma", cp, o_len) == 0)
-				rdma = 1;
-
-			if (_strncmp("nosymlink", cp, o_len) == 0) {
-				warn(ap->logopt, MODPREFIX
-				     "the \"nosymlink\" option is depricated "
-				     "and will soon be removed, "
-				     "use the \"nobind\" option instead");
+			if (strncmp("nosymlink", cp, end - cp + 1) == 0)
 				nosymlink = 1;
-			} else if (_strncmp("nobind", cp, o_len) == 0) {
-				nobind = 1;
-			} else if (_strncmp("no-use-weight-only", cp, o_len) == 0) {
-				flags &= ~MOUNT_FLAG_USE_WEIGHT_ONLY;
-			} else if (_strncmp("use-weight-only", cp, o_len) == 0) {
-				flags |= MOUNT_FLAG_USE_WEIGHT_ONLY;
-			} else {
-				if (_strncmp("vers=4", cp, o_len) == 0 ||
-				    _strncmp("nfsvers=4", cp, o_len) == 0)
-					vers = NFS4_VERS_MASK | TCP_SUPPORTED;
-				else if (_strncmp("vers=3", cp, o_len) == 0 ||
-					 _strncmp("nfsvers=3", cp, o_len) == 0) {
-					vers &= ~(NFS4_VERS_MASK | NFS_VERS_MASK);
-					vers |= NFS3_REQUESTED;
-				} else if (_strncmp("vers=2", cp, o_len) == 0 ||
-					 _strncmp("nfsvers=2", cp, o_len) == 0) {
-					vers &= ~(NFS4_VERS_MASK | NFS_VERS_MASK);
-					vers |= NFS2_REQUESTED;
-				} else if (strstr(cp, "port=") == cp &&
-					 o_len - 5 < 25) {
-					char optport[25];
-
-					strncpy(optport, cp + 5, o_len - 5);
-					optport[o_len - 5] = '\0';
-					port = atoi(optport);
-					if (port < 0)
-						port = 0;
-					port_opt = cp;
-				} else if (_strncmp("proto=udp", cp, o_len) == 0 ||
-					   _strncmp("udp", cp, o_len) == 0) {
-					vers &= ~TCP_SUPPORTED;
-				} else if (_strncmp("proto=tcp", cp, o_len) == 0 ||
-					   _strncmp("tcp", cp, o_len) == 0) {
-					vers &= ~UDP_SUPPORTED;
-				}
+			else {
 				/* Check for options that also make sense
 				   with bind mounts */
-				else if (_strncmp("ro", cp, o_len) == 0)
+				if (strncmp("ro", cp, end - cp + 1) == 0)
 					ro = 1;
 				/* and jump over trailing white space */
 				memcpy(nfsp, cp, comma - cp + 1);
@@ -181,59 +126,22 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 			}
 		}
 
-		/* In case both tcp and udp options were given */
-		if ((vers & NFS_PROTO_MASK) == 0)
-			vers |= NFS_PROTO_MASK;
-
-		debug(ap->logopt, MODPREFIX
-		      "nfs options=\"%s\", nobind=%d, nosymlink=%d, ro=%d",
-		      nfsoptions, nobind, nosymlink, ro);
+		debug(ap->logopt, 
+		      MODPREFIX "nfs options=\"%s\", nosymlink=%d, ro=%d",
+		      nfsoptions, nosymlink, ro);
 	}
 
-	if (!parse_location(ap->logopt, &hosts, what, flags)) {
+	if (strcmp(fstype, "nfs4") == 0)
+		vers = NFS4_VERS_MASK | TCP_SUPPORTED;
+	else
+		vers = NFS_VERS_MASK | NFS_PROTO_MASK;
+
+	if (!parse_location(ap->logopt, &hosts, what)) {
 		info(ap->logopt, MODPREFIX "no hosts available");
 		return 1;
 	}
-	/*
-	 * We can't probe protocol rdma so leave it to mount.nfs(8)
-	 * and and suffer the delay if a server isn't available.
-	 */
-	if (rdma)
-		goto dont_probe;
+	prune_host_list(ap->logopt, &hosts, vers, nfsoptions, random_selection);
 
-	/*
-	 * If this is a singleton mount, and NFSv4 only hasn't been asked
-	 * for, and the default NFS protocol is set to v4 in the autofs
-	 * configuration only probe NFSv4 and let mount.nfs(8) do fallback
-	 * to NFSv3 (if it can). If the NFSv4 probe fails then probe as
-	 * normal.
-	 */
-	if ((hosts && !hosts->next) &&
-	    mount_default_proto == 4 &&
-	    (vers & NFS_VERS_MASK) != 0 &&
-	    (vers & NFS4_VERS_MASK) != 0) {
-		unsigned int v4_probe_ok = 0;
-		struct host *tmp = new_host(hosts->name,
-					    hosts->addr, hosts->addr_len,
-					    hosts->proximity,
-					    hosts->weight, hosts->options);
-		if (tmp) {
-			tmp->rr = hosts->rr;
-			prune_host_list(ap->logopt, &tmp,
-					NFS4_VERS_MASK|TCP_SUPPORTED, port);
-			/* If probe succeeds just try the mount with host in hosts */
-			if (tmp) {
-				v4_probe_ok = 1;
-				free_host_list(&tmp);
-			}
-		}
-		if (!v4_probe_ok)
-			prune_host_list(ap->logopt, &hosts, vers, port);
-	} else {
-		prune_host_list(ap->logopt, &hosts, vers, port);
-	}
-
-dont_probe:
 	if (!hosts) {
 		info(ap->logopt, MODPREFIX "no hosts available");
 		return 1;
@@ -244,10 +152,14 @@ dont_probe:
 	/* Root offset of multi-mount */
 	len = strlen(root);
 	if (root[len - 1] == '/') {
+		fullpath = alloca(len);
 		len = snprintf(fullpath, len, "%s", root);
+	/* Direct mount name is absolute path so don't use root */
 	} else if (*name == '/') {
+		fullpath = alloca(len + 1);
 		len = sprintf(fullpath, "%s", root);
 	} else {
+		fullpath = alloca(len + name_len + 2);
 		len = sprintf(fullpath, "%s/%s", root, name);
 	}
 	fullpath[len] = '\0';
@@ -265,22 +177,21 @@ dont_probe:
 	if (!status)
 		existed = 0;
 
-	/*
-	 * If any *port= option is specified, then we don't want
-	 * a bind mount. Use the "port" option if you want to
-	 * avoid attempting a local bind mount, such as when
-	 * tunneling NFS via localhost.
-	 */
-	if (nfsoptions && *nfsoptions && !port_opt)
-		port_opt = strstr(nfsoptions, "port=");
-
 	this = hosts;
 	while (this) {
-		char *loc;
+		char *loc, *port_opt = NULL;
+
+		/*
+		 * If the "port" option is specified, then we don't want
+		 * a bind mount. Use the "port" option if you want to
+		 * avoid attempting a local bind mount, such as when
+		 * tunneling NFS via localhost.
+		 */
+		if (nfsoptions && *nfsoptions)
+			port_opt = strstr(nfsoptions, "port=");
 
 		/* Port option specified, don't try to bind */
-		if (!(nosymlink || nobind) &&
-		    !port_opt && this->proximity == PROXIMITY_LOCAL) {
+		if (!nosymlink && !port_opt && this->proximity == PROXIMITY_LOCAL) {
 			/* Local host -- do a "bind" */
 			const char *bind_options = ro ? "ro" : "";
 
@@ -311,33 +222,8 @@ dont_probe:
 
 		/* Not a local host - do an NFS mount */
 
-		if (this->rr && this->addr) {
-			socklen_t len = INET6_ADDRSTRLEN;
-			char n_buf[len + 1];
-			const char *n_addr;
-			n_addr = get_addr_string(this->addr, n_buf, len);
-			loc = malloc(strlen(n_addr) + strlen(this->path) + 4);
-			if (!loc) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				error(ap->logopt, "malloc: %s", estr);
-				goto forced_fail;
-			}
-			if (this->addr->sa_family == AF_INET6) {
-				strcpy(loc, "[");
-				strcat(loc, n_addr);
-				strcat(loc, "]");
-			} else
-				strcpy(loc, n_addr);
-		} else {
-			char *host = this->name ? this->name : "localhost";
-			loc = malloc(strlen(host) + strlen(this->path) + 2);
-			if (!loc) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				error(ap->logopt, "malloc: %s", estr);
-				goto forced_fail;
-			}
-			strcpy(loc, host);
-		}
+		loc = malloc(strlen(this->name) + 1 + strlen(this->path) + 1);
+		strcpy(loc, this->name);
 		strcat(loc, ":");
 		strcat(loc, this->path);
 
@@ -358,7 +244,7 @@ dont_probe:
 		}
 
 		if (!err) {
-			debug(ap->logopt, MODPREFIX "mounted %s on %s", loc, fullpath);
+			info(ap->logopt, MODPREFIX "mounted %s on %s", loc, fullpath);
 			free(loc);
 			free_host_list(&hosts);
 			return 0;
