@@ -45,11 +45,15 @@
 #include <stdlib.h>
 #include <sys/errno.h>
 #include <sys/types.h>
-#include <stdint.h>
+#include <netinet/in.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <arpa/inet.h>
+#include <net/if.h>
 #include <netinet/in.h>
 #include <netdb.h>
+#include <unistd.h>
+#include <fcntl.h>
 
 #include "rpc_subs.h"
 #include "replicated.h"
@@ -59,38 +63,217 @@
 #define MAX_ERR_BUF		512
 #endif
 
-#define mymax(x, y)	(x >= y ? x : y)
-#define mmax(x, y, z)	(mymax(x, y) == x ? mymax(x, z) : mymax(y, z))
+#define MAX_IFC_BUF		2048
+static int volatile ifc_buf_len = MAX_IFC_BUF;
+static int volatile ifc_last_len = 0;
+
+#define MASK_A  0x7F000000
+#define MASK_B  0xBFFF0000
+#define MASK_C  0xDFFFFF00
+
+/* Get numeric value of the n bits starting at position p */
+#define getbits(x, p, n)	((x >> (p + 1 - n)) & ~(~0 << n))
+
+#define max(x, y)	(x >= y ? x : y)
+#define mmax(x, y, z)	(max(x, y) == x ? max(x, z) : max(y, z))
 
 void seed_random(void)
 {
 	int fd;
 	unsigned int seed;
 
-	fd = open_fd("/dev/urandom", O_RDONLY);
+	fd = open("/dev/urandom", O_RDONLY);
 	if (fd < 0) {
-		srandom(monotonic_time(NULL));
+		srandom(time(NULL));
 		return;
 	}
 
 	if (read(fd, &seed, sizeof(seed)) != -1)
 		srandom(seed);
 	else
-		srandom(monotonic_time(NULL));
+		srandom(time(NULL));
 
 	close(fd);
 
 	return;
 }
 
-struct host *new_host(const char *name,
-		      struct sockaddr *addr, size_t addr_len,
-		      unsigned int proximity, unsigned int weight,
-		      unsigned int options)
+static int alloc_ifreq(struct ifconf *ifc, int sock)
+{
+	int ret, lastlen = ifc_last_len, len = ifc_buf_len;
+	char err_buf[MAX_ERR_BUF], *buf;
+
+	while (1) {
+		buf = malloc(len);
+		if (!buf) {
+			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
+			logerr("malloc: %s", estr);
+			return 0;
+		}
+
+		ifc->ifc_len = len;
+		ifc->ifc_req = (struct ifreq *) buf;
+
+		ret = ioctl(sock, SIOCGIFCONF, ifc);
+		if (ret == -1) {
+			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
+			logerr("ioctl: %s", estr);
+			free(buf);
+			return 0;
+		}
+
+		if (ifc->ifc_len <= lastlen)
+			break;
+
+		lastlen = ifc->ifc_len;
+		len += MAX_IFC_BUF;
+		free(buf);
+	}
+
+	if (lastlen != ifc_last_len) {
+		ifc_last_len = lastlen;
+		ifc_buf_len = len;
+	}
+
+	return 1;
+}
+
+static unsigned int get_proximity(const char *host_addr, int addr_len)
+{
+	struct sockaddr_in *msk_addr, *if_addr;
+	struct in_addr *hst_addr;
+	char tmp[20], buf[MAX_ERR_BUF], *ptr;
+	struct ifconf ifc;
+	struct ifreq *ifr, nmptr;
+	int sock, cl_flags, ret, i;
+	uint32_t mask, ha, ia;
+
+	memcpy(tmp, host_addr, addr_len);
+	hst_addr = (struct in_addr *) tmp;
+
+	ha = ntohl((uint32_t) hst_addr->s_addr);
+
+	sock = socket(AF_INET, SOCK_DGRAM, 0);
+	if (sock < 0) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr("socket creation failed: %s", estr);
+		return PROXIMITY_ERROR;
+	}
+
+	if ((cl_flags = fcntl(sock, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(sock, F_SETFD, cl_flags);
+	}
+
+	if (!alloc_ifreq(&ifc, sock)) {
+		close(sock);
+		return PROXIMITY_ERROR;
+	}
+
+	/* For each interface */
+
+	/* Is the address a local interface */
+	i = 0;
+	ptr = (char *) &ifc.ifc_buf[0];
+
+	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
+		ifr = (struct ifreq *) ptr;
+
+		switch (ifr->ifr_addr.sa_family) {
+		case AF_INET:
+			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			ret = memcmp(&if_addr->sin_addr, hst_addr, addr_len);
+			if (!ret) {
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_LOCAL;
+			}
+			break;
+
+		default:
+			break;
+		}
+
+		i++;
+		ptr = (char *) &ifc.ifc_req[i];
+	}
+
+	i = 0;
+	ptr = (char *) &ifc.ifc_buf[0];
+
+	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
+		ifr = (struct ifreq *) ptr;
+
+		switch (ifr->ifr_addr.sa_family) {
+		case AF_INET:
+			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			ia =  ntohl((uint32_t) if_addr->sin_addr.s_addr);
+
+			/* Is the address within a localiy attached subnet */
+
+			nmptr = *ifr;
+			ret = ioctl(sock, SIOCGIFNETMASK, &nmptr);
+			if (ret == -1) {
+				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				logerr("ioctl: %s", estr);
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_ERROR;
+			}
+
+			msk_addr = (struct sockaddr_in *) &nmptr.ifr_netmask;
+			mask = ntohl((uint32_t) msk_addr->sin_addr.s_addr);
+
+			if ((ia & mask) == (ha & mask)) {
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_SUBNET;
+			}
+
+			/*
+			 * Is the address within a local ipv4 network.
+			 *
+			 * Bit position 31 == 0 => class A.
+			 * Bit position 30 == 0 => class B.
+			 * Bit position 29 == 0 => class C.
+			 */
+
+			if (!getbits(ia, 31, 1))
+				mask = MASK_A;
+			else if (!getbits(ia, 30, 1))
+				mask = MASK_B;
+			else if (!getbits(ia, 29, 1))
+				mask = MASK_C;
+			else
+				break;
+
+			if ((ia & mask) == (ha & mask)) {
+				close(sock);
+				free(ifc.ifc_req);
+				return PROXIMITY_NET;
+			}
+			break;
+
+		default:
+			break;
+		}
+
+		i++;
+		ptr = (char *) &ifc.ifc_req[i];
+	}
+
+	close(sock);
+	free(ifc.ifc_req);
+
+	return PROXIMITY_OTHER;
+}
+
+static struct host *new_host(const char *name,
+			     const char *addr, size_t addr_len,
+			     unsigned int proximity, unsigned int weight)
 {
 	struct host *new;
-	struct sockaddr *tmp2;
-	char *tmp1;
+	char *tmp1, *tmp2;
 
 	if (!name || !addr)
 		return NULL;
@@ -120,7 +303,6 @@ struct host *new_host(const char *name,
 	new->addr = tmp2;
 	new->proximity = proximity;
 	new->weight = weight;
-	new->options = options;
 
 	return new;
 }
@@ -129,7 +311,7 @@ static int add_host(struct host **list, struct host *host)
 {
 	struct host *this, *last;
 
-	if (!*list) {
+	if (!list) {
 		*list = host;
 		return 1;
 	}
@@ -222,82 +404,77 @@ void free_host_list(struct host **list)
 	*list = NULL;
 }
 
+static unsigned short get_port_option(const char *options)
+{
+	const char *start;
+	long port = 0;
+
+	if (!options)
+		return NFS_PORT;
+
+	start = strstr(options, "port=");
+	if (!start)
+		port = NFS_PORT;
+	else {
+		char optport[30], *opteq, *end;
+		int len;
+
+		end = strchr(start, ',');
+		len = end ? end - start : strlen(start);
+		strncpy(optport, start, len);
+		optport[len] = '\0';
+		opteq = strchr(optport, '=');
+		if (opteq)
+			port = atoi(opteq + 1);
+	}
+
+	if (port < 0)
+		port = 0;
+
+	return (unsigned short) port;
+}
+
 static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 			 struct conn_info *pm_info, struct conn_info *rpc_info,
-			 int proto, unsigned int version, int port)
+			 const char *proto, unsigned int version,
+			 const char *options, unsigned int random_selection)
 {
-	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
-	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
-	socklen_t len = INET6_ADDRSTRLEN;
-	char buf[len + 1];
+	char *have_port_opt = options ? strstr(options, "port=") : NULL;
 	struct pmap parms;
-	struct timespec start, end;
+	struct timeval start, end;
+	struct timezone tz;
 	unsigned int supported = 0;
 	double taken = 0;
 	int status, count = 0;
 
-	if (host->addr)
-		debug(logopt, "called with host %s(%s) proto %d version 0x%x",
-		      host->name, get_addr_string(host->addr, buf, len),
-		      proto, version);
-	else
-		debug(logopt,
-		      "called for host %s proto %d version 0x%x",
-		      host->name, proto, version);
-
-	rpc_info->proto = proto;
-	if (port < 0) {
-		if (version & NFS4_REQUESTED)
-			rpc_info->port = NFS_PORT;
-		else
-			port = 0;
-	} else if (port > 0)
-		rpc_info->port = port;
+	debug(logopt,
+	      "called for host %s proto %s version 0x%x",
+	      host->name, proto, version);
 
 	memset(&parms, 0, sizeof(struct pmap));
+
 	parms.pm_prog = NFS_PROGRAM;
-	parms.pm_prot = proto;
+
+	/* Try to prode UDP first to conserve socket space */
+	rpc_info->proto = getprotobyname(proto);
+	if (!rpc_info->proto)
+		return 0;
 
 	if (!(version & NFS4_REQUESTED))
 		goto v3_ver;
 
-	if (!port) {
-		status = rpc_portmap_getclient(pm_info,
-				host->name, host->addr, host->addr_len,
-				proto, RPC_CLOSE_DEFAULT);
-		if (status == -EHOSTUNREACH) {
-			supported = status;
-			goto done_ver;
-		} else if (status)
-			goto done_ver;
-		parms.pm_vers = NFS4_VERSION;
-		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
-		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
-			supported = status;
-			goto done_ver;
-		} else if (status < 0) {
-			if (version & NFS_VERS_MASK)
-				goto v3_ver; /* MOUNT_NFS_DEFAULT_PROTOCOL=4 */
-			else
-				goto done_ver;
-		}
-	}
+	if (!(rpc_info->port = get_port_option(options)))
+		goto v3_ver;
 
-	if (rpc_info->proto == IPPROTO_UDP)
+	if (rpc_info->proto->p_proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
-	if (status == -EHOSTUNREACH) {
-		supported = status;
-		goto done_ver;
-	} else if (!status) {
-		clock_gettime(CLOCK_MONOTONIC, &start);
+	if (status) {
+		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
-		clock_gettime(CLOCK_MONOTONIC, &end);
-		if (status == -ETIMEDOUT) {
-			supported = status;
-			goto done_ver;
-		} else if (status > 0) {
+		gettimeofday(&end, &tz);
+		if (status) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -305,7 +482,7 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 				debug(logopt,
 				      "nfs v4 random selection time: %f", reply);
 			} else {
-				reply = monotonic_elapsed(start, end);
+				reply = elapsed(start, end);
 				debug(logopt, "nfs v4 rpc ping time: %f", reply);
 			}
 			taken += reply;
@@ -314,49 +491,38 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 		}
 	}
 
-	if (!(version & NFS_VERS_MASK))
-		goto done_ver;
-
 v3_ver:
-	if (!(version & NFS3_REQUESTED))
-		goto v2_ver;
-
-	if (!port && !pm_info->client) {
+	if (!have_port_opt) {
 		status = rpc_portmap_getclient(pm_info,
 				host->name, host->addr, host->addr_len,
 				proto, RPC_CLOSE_DEFAULT);
-		if (status == -EHOSTUNREACH) {
-			supported = status;
-			goto done_ver;
-		} else if (status)
+		if (!status)
 			goto done_ver;
 	}
 
-	if (!port) {
-		parms.pm_vers = NFS3_VERSION;
-		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
-		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
-			supported = status;
+	if (!(version & NFS3_REQUESTED))
+		goto v2_ver;
+
+	if (have_port_opt) {
+		if (!(rpc_info->port = get_port_option(options)))
 			goto done_ver;
-		} else if (status < 0)
+	} else {
+		parms.pm_prot = rpc_info->proto->p_proto;
+		parms.pm_vers = NFS3_VERSION;
+		rpc_info->port = rpc_portmap_getport(pm_info, &parms);
+		if (!rpc_info->port)
 			goto v2_ver;
 	}
 
-	if (rpc_info->proto == IPPROTO_UDP)
+	if (rpc_info->proto->p_proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
-	if (status == -EHOSTUNREACH) {
-		supported = status;
-		goto done_ver;
-	} else if (!status) {
-		clock_gettime(CLOCK_MONOTONIC, &start);
+	if (status) {
+		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
-		clock_gettime(CLOCK_MONOTONIC, &end);
-		if (status == -ETIMEDOUT) {
-			supported = status;
-			goto done_ver;
-		} else if (status > 0) {
+		gettimeofday(&end, &tz);
+		if (status) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -364,7 +530,7 @@ v3_ver:
 				debug(logopt,
 				      "nfs v3 random selection time: %f", reply);
 			} else {
-				reply = monotonic_elapsed(start, end);
+				reply = elapsed(start, end);
 				debug(logopt, "nfs v3 rpc ping time: %f", reply);
 			}
 			taken += reply;
@@ -377,41 +543,26 @@ v2_ver:
 	if (!(version & NFS2_REQUESTED))
 		goto done_ver;
 
-	if (!port && !pm_info->client) {
-		status = rpc_portmap_getclient(pm_info,
-				host->name, host->addr, host->addr_len,
-				proto, RPC_CLOSE_DEFAULT);
-		if (status == -EHOSTUNREACH) {
-			supported = status;
+	if (have_port_opt) {
+		if (!(rpc_info->port = get_port_option(options)))
 			goto done_ver;
-		} else if (status)
-			goto done_ver;
-	}
-
-	if (!port) {
+	} else {
+		parms.pm_prot = rpc_info->proto->p_proto;
 		parms.pm_vers = NFS2_VERSION;
-		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
-		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
-			supported = status;
-			goto done_ver;
-		} else if (status < 0)
+		rpc_info->port = rpc_portmap_getport(pm_info, &parms);
+		if (!rpc_info->port)
 			goto done_ver;
 	}
 
-	if (rpc_info->proto == IPPROTO_UDP)
+	if (rpc_info->proto->p_proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
-	if (status == -EHOSTUNREACH) {
-		supported = status;
-		goto done_ver;
-	} else if (!status) {
-		clock_gettime(CLOCK_MONOTONIC, &start);
+	if (status) {
+		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(rpc_info);
-		clock_gettime(CLOCK_MONOTONIC, &end);
-		if (status == -ETIMEDOUT)
-			supported = status;
-		else if (status > 0) {
+		gettimeofday(&end, &tz);
+		if (status) {
 			double reply;
 			if (random_selection) {
 				/* Random value between 0 and 1 */
@@ -419,7 +570,7 @@ v2_ver:
 				debug(logopt,
 				      "nfs v2 random selection time: %f", reply);
 			} else {
-				reply = monotonic_elapsed(start, end);;
+				reply = elapsed(start, end);;
 				debug(logopt, "nfs v2 rpc ping time: %f", reply);
 			}
 			taken += reply;
@@ -429,7 +580,7 @@ v2_ver:
 	}
 
 done_ver:
-	if (rpc_info->proto == IPPROTO_UDP) {
+	if (rpc_info->proto->p_proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(rpc_info);
 		rpc_destroy_udp_client(pm_info);
 	} else {
@@ -442,10 +593,7 @@ done_ver:
 		 * Average response time to 7 significant places as
 		 * integral type.
 		 */
-		if (use_weight_only)
-			host->cost = 1;
-		else
-			host->cost = (unsigned long) ((taken * 1000000) / count);
+		host->cost = (unsigned long) ((taken * 1000000) / count);
 
 		/* Allow for user bias */
 		if (host->weight)
@@ -459,7 +607,8 @@ done_ver:
 }
 
 static int get_vers_and_cost(unsigned logopt, struct host *host,
-			     unsigned int version, int port)
+			     unsigned int version, const char *options,
+			     unsigned int random_selection)
 {
 	struct conn_info pm_info, rpc_info;
 	time_t timeout = RPC_TIMEOUT;
@@ -484,28 +633,23 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	vers &= version;
 
-	if (version & TCP_REQUESTED) {
+	if (version & UDP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-				   &pm_info, &rpc_info, IPPROTO_TCP, vers, port);
-		if (IS_ERR(supported)) {
-			if (ERR(supported) == EHOSTUNREACH ||
-			    ERR(supported) == ETIMEDOUT)
-				return ret;
-		} else if (supported) {
+					&pm_info, &rpc_info, "udp", vers,
+					options, random_selection);
+		if (supported) {
 			ret = 1;
-			host->version |= supported;
+			host->version |= (supported << 8);
 		}
 	}
 
-	if (version & UDP_REQUESTED) {
+	if (version & TCP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-				   &pm_info, &rpc_info, IPPROTO_UDP, vers, port);
-		if (IS_ERR(supported)) {
-			if (!ret && ERR(supported) == ETIMEDOUT)
-				return ret;
-		} else if (supported) {
+					 &pm_info, &rpc_info, "tcp", vers,
+					 options, random_selection);
+		if (supported) {
 			ret = 1;
-			host->version |= (supported << 8);
+			host->version |= supported;
 		}
 	}
 
@@ -513,30 +657,26 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 }
 
 static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
-				      unsigned int version, int port)
+				      unsigned int version, const char *options,
+				      unsigned int random_selection)
 {
-	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
-	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
-	socklen_t len = INET6_ADDRSTRLEN;
-	char buf[len + 1];
+	char *have_port_opt = options ? strstr(options, "port=") : NULL;
 	struct conn_info pm_info, rpc_info;
-	int proto;
+	struct pmap parms;
+	const char *proto;
 	unsigned int vers;
-	struct timespec start, end;
+	struct timeval start, end;
+	struct timezone tz;
 	double taken = 0;
 	time_t timeout = RPC_TIMEOUT;
-	int status = 0;
+	int status;
 
-	if (host->addr)
-		debug(logopt, "called with host %s(%s) version 0x%x",
-			host->name, get_addr_string(host->addr, buf, len),
-			version);
-	else
-		debug(logopt, "called with host %s version 0x%x",
-			host->name, version);
+	debug(logopt,
+	      "called with host %s version 0x%x", host->name, version);
 
 	memset(&pm_info, 0, sizeof(struct conn_info));
 	memset(&rpc_info, 0, sizeof(struct conn_info));
+	memset(&parms, 0, sizeof(struct pmap));
 
 	if (host->proximity == PROXIMITY_NET)
 		timeout = RPC_TIMEOUT * 2;
@@ -551,6 +691,8 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	rpc_info.close_option = RPC_CLOSE_DEFAULT;
 	rpc_info.client = NULL;
 
+	parms.pm_prog = NFS_PROGRAM;
+
 	/*
 	 *  The version passed in is the version as defined in
 	 *  include/replicated.h.  However, the version we want to send
@@ -558,10 +700,10 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	 *  So, we do the conversion here.
 	 */
 	if (version & UDP_SELECTED_MASK) {
-		proto = IPPROTO_UDP;
+		proto = "udp";
 		version >>= 8;
 	} else
-		proto = IPPROTO_TCP;
+		proto = "tcp";
 
 	switch (version) {
 	case NFS2_SUPPORTED:
@@ -578,52 +720,50 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		return 0;
 	}
 
-	rpc_info.proto = proto;
+	rpc_info.proto = getprotobyname(proto);
+	if (!rpc_info.proto)
+		return 0;
 
-	if (port > 0)
-		rpc_info.port = port;
-	else if (vers & NFS4_VERSION && port < 0)
-		rpc_info.port = NFS_PORT;
-	else {
-		struct pmap parms;
+	status = 0;
+
+	parms.pm_vers = vers;
+	if (have_port_opt || (vers & NFS4_VERSION)) {
+		if (!(rpc_info.port = get_port_option(options)))
+			return 0;
+	} else {
 		int ret = rpc_portmap_getclient(&pm_info,
 				host->name, host->addr, host->addr_len,
 				proto, RPC_CLOSE_DEFAULT);
-		if (ret)
+		if (!ret)
 			return 0;
 
-		memset(&parms, 0, sizeof(struct pmap));
-		parms.pm_prog = NFS_PROGRAM;
-		parms.pm_prot = rpc_info.proto;
-		parms.pm_vers = vers;
-		ret = rpc_portmap_getport(&pm_info, &parms, &rpc_info.port);
-		if (ret < 0)
+		parms.pm_prot = rpc_info.proto->p_proto;
+		rpc_info.port = rpc_portmap_getport(&pm_info, &parms);
+		if (!rpc_info.port)
 			goto done;
 	}
 
-	if (rpc_info.proto == IPPROTO_UDP)
-		status = rpc_udp_getclient(&rpc_info, NFS_PROGRAM, vers);
+	if (rpc_info.proto->p_proto == IPPROTO_UDP)
+		status = rpc_udp_getclient(&rpc_info, NFS_PROGRAM, parms.pm_vers);
 	else
-		status = rpc_tcp_getclient(&rpc_info, NFS_PROGRAM, vers);
-	if (status == -EHOSTUNREACH)
-		goto done;
-	else if (!status) {
-		clock_gettime(CLOCK_MONOTONIC, &start);
+		status = rpc_tcp_getclient(&rpc_info, NFS_PROGRAM, parms.pm_vers);
+	if (status) {
+		gettimeofday(&start, &tz);
 		status = rpc_ping_proto(&rpc_info);
-		clock_gettime(CLOCK_MONOTONIC, &end);
-		if (status > 0) {
+		gettimeofday(&end, &tz);
+		if (status) {
 			if (random_selection) {
 				/* Random value between 0 and 1 */
 				taken = ((float) random())/((float) RAND_MAX+1);
 				debug(logopt, "random selection time %f", taken);
 			} else {
-				taken = monotonic_elapsed(start, end);
+				taken = elapsed(start, end);
 				debug(logopt, "rpc ping time %f", taken);
 			}
 		}
 	}
 done:
-	if (rpc_info.proto == IPPROTO_UDP) {
+	if (rpc_info.proto->p_proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(&rpc_info);
 		rpc_destroy_udp_client(&pm_info);
 	} else {
@@ -633,10 +773,7 @@ done:
 
 	if (status) {
 		/* Response time to 7 significant places as integral type. */
-		if (use_weight_only)
-			host->cost = 1;
-		else
-			host->cost = (unsigned long) (taken * 1000000);
+		host->cost = (unsigned long) (taken * 1000000);
 
 		/* Allow for user bias */
 		if (host->weight)
@@ -651,7 +788,8 @@ done:
 }
 
 int prune_host_list(unsigned logopt, struct host **list,
-		    unsigned int vers, int port)
+		    unsigned int vers, const char *options,
+		    unsigned int random_selection)
 {
 	struct host *this, *last, *first;
 	struct host *new = NULL;
@@ -660,17 +798,10 @@ int prune_host_list(unsigned logopt, struct host **list,
 	unsigned int v2_udp_count, v3_udp_count, v4_udp_count;
 	unsigned int max_udp_count, max_tcp_count, max_count;
 	int status;
-	int kern_vers;
 
 	if (!*list)
 		return 0;
 
-	/* If we're using the host name then there's no point probing
-	 * avialability and respose time.
-	 */
-	if (defaults_use_hostname_for_mounts())
-		return 1;
-
 	/* Use closest hosts to choose NFS version */
 
 	first = *list;
@@ -679,34 +810,19 @@ int prune_host_list(unsigned logopt, struct host **list,
 	this = first;
 	while (this && this->proximity == PROXIMITY_LOCAL)
 		this = this->next;
-	first = this;
 
 	/*
 	 * Check for either a list containing only proximity local hosts
 	 * or a single host entry whose proximity isn't local. If so
 	 * return immediately as we don't want to add probe latency for
 	 * the common case of a single filesystem mount request.
-	 *
-	 * But, if the kernel understands text nfs mount options then
-	 * mount.nfs most likely bypasses its probing and lets the kernel
-	 * do all the work. This can lead to long timeouts for hosts that
-	 * are not available so check the kernel version and mount.nfs
-	 * version and probe singleton mounts if the kernel version is
-	 * greater than 2.6.22 and mount.nfs version is greater than 1.1.1.
-	 * But also allow the MOUNT_WAIT configuration parameter to override
-	 * the probing.
 	 */
-	if (nfs_mount_uses_string_options &&
-	    defaults_get_mount_wait() == -1 &&
-	   (kern_vers = linux_version_code()) > KERNEL_VERSION(2, 6, 22)) {
-		if (!this)
-			return 1;
-	} else {
-		if (!this || !this->next)
-			return 1;
-	}
+	if (!this || !this->next)
+		return 1;
 
 	proximity = this->proximity;
+	first = this;
+	this = first;
 	while (this) {
 		struct host *next = this->next;
 
@@ -714,7 +830,8 @@ int prune_host_list(unsigned logopt, struct host **list,
 			break;
 
 		if (this->name) {
-			status = get_vers_and_cost(logopt, this, vers, port);
+			status = get_vers_and_cost(logopt, this, vers,
+						   options, random_selection);
 			if (!status) {
 				if (this == first) {
 					first = next;
@@ -768,7 +885,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 
 	max_tcp_count = mmax(v4_tcp_count, v3_tcp_count, v2_tcp_count);
 	max_udp_count = mmax(v4_udp_count, v3_udp_count, v2_udp_count);
-	max_count = mymax(max_tcp_count, max_udp_count);
+	max_count = max(max_tcp_count, max_udp_count);
 
 	if (max_count == v4_tcp_count) {
 		selected_version = NFS4_TCP_SUPPORTED;
@@ -823,7 +940,8 @@ int prune_host_list(unsigned logopt, struct host **list,
 			add_host(&new, this);
 		} else {
 			status = get_supported_ver_and_cost(logopt, this,
-						selected_version, port);
+						selected_version, options,
+						random_selection);
 			if (status) {
 				this->version = selected_version;
 				remove_host(list, this);
@@ -839,144 +957,67 @@ int prune_host_list(unsigned logopt, struct host **list,
 	return 1;
 }
 
-static int add_new_host(struct host **list,
-			const char *host, unsigned int weight,
-			struct addrinfo *host_addr,
-			unsigned int rr, unsigned int options)
+static int add_host_addrs(struct host **list, const char *host, unsigned int weight)
 {
+	struct hostent he;
+	struct hostent *phe = &he;
+	struct hostent *result;
+	struct sockaddr_in saddr;
+	char buf[MAX_IFC_BUF], **haddr;
+	int ghn_errno, ret;
 	struct host *new;
 	unsigned int prx;
-	int addr_len;
 
-	prx = get_proximity(host_addr->ai_addr);
+	saddr.sin_family = AF_INET;
+	if (inet_aton(host, &saddr.sin_addr)) {
+		const char *thost = (const char *) &saddr.sin_addr;
 
-	/*
-	 * If we want the weight to be the determining factor
-	 * when selecting a host, or we are using random selection,
-	 * then all hosts must have the same proximity. However,
-	 * if this is the local machine it should always be used
-	 * since it is certainly available.
-	 */
-	if (prx != PROXIMITY_LOCAL &&
-	   (options & (MOUNT_FLAG_USE_WEIGHT_ONLY |
-		       MOUNT_FLAG_RANDOM_SELECT)))
-		prx = PROXIMITY_SUBNET;
+		prx = get_proximity(thost, sizeof(saddr.sin_addr));
+		if (prx == PROXIMITY_ERROR)
+			return 0;
+
+		if (!(new = new_host(host, thost, sizeof(saddr.sin_addr), prx, weight)))
+			return 0;
+
+		if (!add_host(list, new))
+			free_host(new);
 
-	/*
-	 * If we tried to add an IPv6 address and we don't have IPv6
-	 * support return success in the hope of getting an IPv4
-	 * address later.
-	 */
-	if (prx == PROXIMITY_UNSUPPORTED)
 		return 1;
-	if (prx == PROXIMITY_ERROR)
-		return 0;
+	}
 
-	if (host_addr->ai_addr->sa_family == AF_INET)
-		addr_len = INET_ADDRSTRLEN;
-	else if (host_addr->ai_addr->sa_family == AF_INET6)
-		addr_len = INET6_ADDRSTRLEN;
-	else
-		return 0;
+	memset(buf, 0, MAX_IFC_BUF);
+	memset(&he, 0, sizeof(struct hostent));
 
-	new = new_host(host, host_addr->ai_addr, addr_len, prx, weight, options);
-	if (!new)
-		return 0;
-
-	if (!add_host(list, new)) {
-		free_host(new);
+	ret = gethostbyname_r(host, phe,
+			buf, MAX_IFC_BUF, &result, &ghn_errno);
+	if (ret || !result) {
+		if (ghn_errno == -1)
+			logmsg("host %s: lookup failure %d", host, errno);
+		else
+			logmsg("host %s: lookup failure %d", host, ghn_errno);
 		return 0;
 	}
-	new->rr = rr;
+
+	for (haddr = phe->h_addr_list; *haddr; haddr++) {
+		struct in_addr tt;
+
+		prx = get_proximity(*haddr, phe->h_length);
+		if (prx == PROXIMITY_ERROR)
+			return 0;
+
+		memcpy(&tt, *haddr, sizeof(struct in_addr));
+		if (!(new = new_host(host, *haddr, phe->h_length, prx, weight)))
+			return 0;
+
+		if (!add_host(list, new)) {
+			free_host(new);
+			continue;
+		}
+	}
 
 	return 1;
 }
 
-static int add_host_addrs(struct host **list, const char *host,
-			  unsigned int weight, unsigned int options)
-{
-	struct addrinfo hints, *ni, *this;
-	char *n_ptr;
-	char *name = n_ptr = strdup(host);
-	int len;
-	char buf[MAX_ERR_BUF];
-	int rr = 0, rr4 = 0, rr6 = 0;
-	int ret;
-
-	if (!name) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(LOGOPT_ANY, "strdup: %s", estr);
-		error(LOGOPT_ANY, "failed to add host %s", host);
-		return 0;
-	}
-	len = strlen(name);
-
-	if (name[0] == '[' && name[--len] == ']') {
-		name[len] = '\0';
-		name++;
-	}
-
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_flags = AI_NUMERICHOST;
-	hints.ai_family = AF_UNSPEC;
-	hints.ai_socktype = SOCK_DGRAM;
-
-	ret = getaddrinfo(name, NULL, &hints, &ni);
-	if (ret)
-		goto try_name;
-
-	this = ni;
-	while (this) {
-		ret = add_new_host(list, host, weight, this, 0, options);
-		if (!ret)
-			break;
-		this = this->ai_next;
-	}
-	freeaddrinfo(ni);
-	goto done;
-
-try_name:
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_flags = AI_ADDRCONFIG;
-	hints.ai_family = AF_UNSPEC;
-	hints.ai_socktype = SOCK_DGRAM;
-
-	ret = getaddrinfo(name, NULL, &hints, &ni);
-	if (ret) {
-		error(LOGOPT_ANY, "hostname lookup failed: %s",
-		      gai_strerror(ret));
-		free(name);
-		return 0;
-	}
-
-	this = ni;
-	while (this) {
-		if (this->ai_family == AF_INET) {
-			struct sockaddr_in *addr = (struct sockaddr_in *) this->ai_addr;
-			if (addr->sin_addr.s_addr != INADDR_LOOPBACK)
-				rr4++;
-		} else if (this->ai_family == AF_INET6) {
-			struct sockaddr_in6 *addr = (struct sockaddr_in6 *) this->ai_addr;
-			if (!IN6_IS_ADDR_LOOPBACK(addr->sin6_addr.s6_addr32))
-				rr6++;
-		}
-		this = this->ai_next;
-	}
-	if (rr4 > 1 || rr6 > 1)
-		rr++;
-	this = ni;
-	while (this) {
-		ret = add_new_host(list, host, weight, this, rr, options);
-		if (!ret)
-			break;
-		this = this->ai_next;
-	}
-	freeaddrinfo(ni);
-done:
-	free(n_ptr);
-	return ret;
-}
-
 static int add_path(struct host *hosts, const char *path, int len)
 {
 	struct host *this;
@@ -1023,8 +1064,7 @@ static int add_local_path(struct host **hosts, const char *path)
 	new->path = tmp;
 	new->proximity = PROXIMITY_LOCAL;
 	new->version = NFS_VERS_MASK;
-	new->name = NULL;
-	new->addr = NULL;
+	new->name = new->addr = NULL;
 	new->weight = new->cost = 0;
 
 	add_host(hosts, new);
@@ -1032,30 +1072,7 @@ static int add_local_path(struct host **hosts, const char *path)
 	return 1;
 }
 
-static char *seek_delim(const char *s)
-{
-	const char *p = s;
-	char *delim;
-
-	delim = strpbrk(p, "(, \t:");
-	if (delim && *delim != ':' && (delim == s || *(delim - 1) != '\\'))
-		return delim;
-
-	while (*p) {
-		if (*p != ':') {
-			p++;
-			continue;
-		}
-		if (!strncmp(p, ":/", 2))
-			return (char *) p;
-		p++;
-	}
-
-	return NULL;
-}
-
-int parse_location(unsigned logopt, struct host **hosts,
-		   const char *list, unsigned int options)
+int parse_location(unsigned logopt, struct host **hosts, const char *list)
 {
 	char *str, *p, *delim;
 	unsigned int empty = 1;
@@ -1074,7 +1091,7 @@ int parse_location(unsigned logopt, struct host **hosts,
 		int weight = 0;
 
 		p += strspn(p, " \t,");
-		delim = seek_delim(p);
+		delim = strpbrk(p, "(, \t:");
 
 		if (delim) {
 			if (*delim == '(') {
@@ -1087,12 +1104,6 @@ int parse_location(unsigned logopt, struct host **hosts,
 					*delim = '\0';
 					weight = atoi(w);
 				}
-				else {
-					/* syntax error - Mismatched brackets */
-					free_host_list(hosts);
-					free(str);
-					return 0;
-				}
 				delim++;
 			}
 
@@ -1104,7 +1115,7 @@ int parse_location(unsigned logopt, struct host **hosts,
 
 				/* Oh boy - might have spaces in the path */
 				next = path;
-				while (*next && strncmp(next, ":/", 2))
+				while (*next && *next != ':')
 					next++;
 
 				/* No spaces in host names at least */
@@ -1116,7 +1127,7 @@ int parse_location(unsigned logopt, struct host **hosts,
 				}
 
 				if (p != delim) {
-					if (!add_host_addrs(hosts, p, weight, options)) {
+					if (!add_host_addrs(hosts, p, weight)) {
 						if (empty) {
 							p = next;
 							continue;
@@ -1138,7 +1149,7 @@ int parse_location(unsigned logopt, struct host **hosts,
 				*delim = '\0';
 				next = delim + 1;
 
-				if (!add_host_addrs(hosts, p, weight, options)) {
+				if (!add_host_addrs(hosts, p, weight)) {
 					p = next;
 					continue;
 				}
